"""Testing rq"""

import pytest
import pandas as pd
from pdpipe import rq


#     a   b
# 1   1   4
# 2   4   5
# 3  18  11
def _df1():
    return pd.DataFrame([[1, 4], [4, 5], [18, 11]], [1, 2, 3], ['a', 'b'])


#        a   b
# 1      1   4
# 2   Null   5
# 3     18  11
def _df2():
    return pd.DataFrame([[1, 4], [None, 5], [18, 11]], [1, 2, 3], ['a', 'b'])


def test_rq():
    df = _df1()
    res = rq.ColValGt('a', 4)(df)
    assert res.eq([False, False, True]).all()
    res = rq.ColValGe('a', 4)(df)
    assert res.eq([False, True, True]).all()
    res = rq.ColValLt('a', 4)(df)
    assert res.eq([True, False, False]).all()
    res = rq.ColValLe('a', 4)(df)
    assert res.eq([True, True, False]).all()
    res = rq.ColValEq('a', 4)(df)
    assert res.eq([False, True, False]).all()
    res = rq.ColValNe('a', 4)(df)
    assert res.eq([True, False, True]).all()
    res = rq.ColValIsIn('a', [1, 2, 3, 4])(df)
    assert res.eq([True, True, False]).all()
    df2 = _df2()
    res = rq.ColValIsNa('a')(df2)
    assert res.eq([False, True, False]).all()
    res = rq.ColValNotNa('a')(df2)
    assert res.eq([True, False, True]).all()

    # unary operators
    res = (~ rq.ColValLe('a', 4))(df)
    assert res.eq([False, False, True]).all()

    # binary operators
    res = (rq.ColValLe('a', 1) | rq.ColValGe('b', 11))(df)
    assert res.eq([True, False, True]).all()
    res = (rq.ColValLe('a', 1) & rq.ColValGe('b', 11))(df)
    assert res.eq([False, False, False]).all()
    res = (rq.ColValLe('a', 4) | rq.ColValGe('b', 5))(df)
    assert res.eq([True, True, True]).all()
    res = (rq.ColValLe('a', 4) ^ rq.ColValGe('b', 5))(df)
    assert res.eq([True, False, True]).all()

    # doc-less func
    q = rq.RowQualifier(lambda df: df['a'] > 4)
    print(q)

    # errors
    q = rq.ColValGe('a', 4)
    with pytest.raises(TypeError):
        q & '4'
    with pytest.raises(TypeError):
        q | '4'
    with pytest.raises(TypeError):
        q ^ '4'



"""Test the ColumnDtypeEnforcer pipeline stage."""

import pandas as pd

from pdpipe import ColumnDtypeEnforcer
import pdpipe.cq as cq


DF = pd.DataFrame([[8, 'a'], [5, 'b']], [1, 2], ['num', 'initial'])


def test_dtype_enf_basic():
    stage = ColumnDtypeEnforcer({'num': float})
    res = stage(DF)
    assert res['num'].dtype == float


DF2 = pd.DataFrame(
    [[8, 2, 0], [5, 1, 1], [1, 2, 1]], [1, 2, 3], ['num1', 'num2', 'rank'])


def test_dtype_enf_col_qualifier():
    stage = ColumnDtypeEnforcer({'rank': bool, cq.StartWith('num'): float})
    res = stage(DF2)
    assert res['num1'].dtype == float
    assert res['num2'].dtype == float
    assert res['rank'].dtype == bool

    # Only col_qualifier as key, used as documentation example
    res = ColumnDtypeEnforcer({cq.StartWith('n'): float}).apply(DF)
    assert res['num'].dtype == float



"""Column qualifiers for pdpipe."""

import numpy as np

from .shared import _list_str


class UnfittedColumnQualifierError(Exception):
    """An exception raised when a (non-fit) transform is attempted with an
    unfitted column qualifier.
    """


class ColumnQualifier(object):
    """A fittable qualifier that returns column labels from an input dataframe.

    Parameters
    ----------
    func : callable
        A callable that given an input pandas.DataFrame objects returns a list
        of labels of a subset of the columns of the input dataframe.
    fittable : bool, default True
        If set to false, this qualifier becomes unfittable, and `func` is
        called on every call to transform. True by default.
    subset : bool, default False
        If set to true, fitted qualifiers return the subset of fitted columns
        found in input dataframes during transform, in the order they appeared
        when fitted (NOT in the order they appear in the input dataframe of the
        transform). False by default, which means fitted qualifiers return the
        FULL list of fitted columns, ignoring input dataframes completely on
        transforms. When combined with most pipeline stages, this means the
        stage will fail on its precondition if trying to transform with it a
        dataframe that is missing some values in the fitted qualifier.

    Example
    -------
    >>> import numpy as np; import pdpipe as pdp;
    >>> cq = pdp.cq.ColumnQualifier(lambda df: [
    ...    l for l, s in df.iteritems()
    ...    if s.dtype == np.int64 and l in ['a', 'b', 5]
    ... ])
    >>> cq
    <ColumnQualifier: Qualify columns by function>
    >>> col_drop = pdp.ColDrop(columns=cq)
    """

    def __init__(self, func, fittable=None, subset=None):
        if fittable is None:
            fittable = True
        self._cqfunc = func
        self.__doc__ = func.__doc__
        self._fittable = fittable
        self._subset = subset

    def __call__(self, X):
        """Returns column labels of qualified columns from an input dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The input dataframe, from which columns are selected.

        Returns
        -------
        list of objects
            A list of labels of the qualified columns for the input dataframe.
        """
        try:
            return self.transform(X)
        except UnfittedColumnQualifierError:
            return self.fit_transform(X)

    def fit_transform(self, X):
        """Fits this qualifier and returns the labels of the qualifying columns.

        Parameters
        ----------
        X : pandas.DataFrame
            The input dataframe, from which columns are selected.

        Returns
        -------
        list of objects
            A list of labels of the qualified columns for the input dataframe.
        """
        self._columns = self._cqfunc(X)
        return self._columns

    def fit(self, X):
        """Fits this qualifier on the input dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The input dataframe, from which columns are selected.

        """
        self.fit_transform(X)

    def transform(self, X):
        """Applies and returns the labels of the qualifying columns.

        If this ColumnQualifier is fittable, it will return the list of column
        labels that was determined when fitted (or the subset of it that can
        be found in the input dataframe). It will throw an exception if it
        is not.

        Parameters
        ----------
        X : pandas.DataFrame
            The input dataframe, from which columns are selected.

        Returns
        -------
        list of objects
            A list of labels of the qualified columns for the input dataframe.
        """
        if not self._fittable:
            return self._cqfunc(X)
        try:
            if self._subset:
                return [x for x in self._columns if x in X.columns]
            return self._columns
        except AttributeError:
            raise UnfittedColumnQualifierError

    def __repr__(self):
        fstr = ''
        if self._cqfunc.__doc__:  # pragma: no cover
            fstr = f' - {self._cqfunc.__doc__}'
        return f"<ColumnQualifier: Qualify columns by function{fstr}>"

    # --- overriding boolean operators ---

    @staticmethod
    def _x_inorderof_y(x, y):
        return [i for i in y if i in x]

    class _AndQualifierFunc(object):
        """A pickle-able AND qualifier class."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, X):
            return ColumnQualifier._x_inorderof_y(
                x=set(self.first(X)).intersection(self.second(X)),
                y=X.columns,
            )

    def __and__(self, other):
        try:
            res_func = ColumnQualifier._AndQualifierFunc(
                first=self._cqfunc,
                second=other._cqfunc,
            )
            res_func.__doc__ = (
                f"{self._cqfunc.__doc__ or 'Anonymous qualifier 1'} AND "
                f"{other._cqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return ColumnQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _XorQualifierFunc(object):
        """A pickle-able XOR qualifier class."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, X):
            return ColumnQualifier._x_inorderof_y(
                x=set(self.first(X)).symmetric_difference(self.second(X)),
                y=X.columns,
            )

    def __xor__(self, other):
        try:
            res_func = ColumnQualifier._XorQualifierFunc(
                first=self._cqfunc,
                second=other._cqfunc,
            )
            res_func.__doc__ = (
                f"{self._cqfunc.__doc__ or 'Anonymous qualifier 1'} XOR "
                f"{other._cqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return ColumnQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _OrQualifierFunc(object):
        """A pickle-able OR qualifier class."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, X):
            return ColumnQualifier._x_inorderof_y(
                x=set(self.first(X)).union(self.second(X)),
                y=X.columns,
            )

    def __or__(self, other):
        try:
            res_func = ColumnQualifier._OrQualifierFunc(
                first=self._cqfunc,
                second=other._cqfunc,
            )
            res_func.__doc__ = (
                f"{self._cqfunc.__doc__ or 'Anonymous qualifier 1'} OR "
                f"{other._cqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return ColumnQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _SubQualifierFunc(object):
        """A pickle-able SUB qualifier class."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, X):
            return ColumnQualifier._x_inorderof_y(
                x=set(self.first(X)).difference(self.second(X)),
                y=X.columns,
            )

    def __sub__(self, other):
        try:
            res_func = ColumnQualifier._SubQualifierFunc(
                first=self._cqfunc,
                second=other._cqfunc,
            )
            res_func.__doc__ = (
                f"{self._cqfunc.__doc__ or 'Anonymous qualifier 1'} NOT IN "
                f"{other._cqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return ColumnQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _NotQualifierFunc(object):
        """A pickle-able NOT qualifier class."""

        def __init__(self, cq):
            self.cq = cq

        def __call__(self, X):
            return ColumnQualifier._x_inorderof_y(
                x=set(X.columns).difference(self.cq(X)),
                y=X.columns,
            )

    def __invert__(self):
        res_func = ColumnQualifier._NotQualifierFunc(
            cq=self._cqfunc
        )
        res_func.__doc__ = (
            f"NOT {self._cqfunc.__doc__ or 'Anonymous qualifier'}"
        )
        return ColumnQualifier(func=res_func)


def is_fittable_column_qualifier(obj):
    """Returns True for objects that are fittable ColumnQualifier objects.

    Parameters
    ----------
    obj : object
        The object to examine.

    Returns
    -------
    bool
        True if the given object is an instance of ColumnQualifier and
        fittable, False otherwise.
    """
    return isinstance(obj, ColumnQualifier) and obj._fittable


class AllColumns(ColumnQualifier):
    """Selects all columns in input dataframes.

    Parameters
    ----------
    **kwargs
        Accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of `ColumnQualifier` for
        details.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[8,1],[5,2]], [1,2], ['a', 'b'])
    >>> cq = pdp.cq.AllColumns()
    >>> cq
    <ColumnQualifier: Qualify all columns>
    >>> cq(df)
    ['a', 'b']
    >>> df2 = pd.DataFrame([[8,1],[5,2]], [1,2], ['b', 'c'])
    >>> cq(df2)
    ['a', 'b']
    >>> cq = pdp.cq.AllColumns(fittable=False)
    >>> cq(df)
    ['a', 'b']
    >>> cq(df2)
    ['b', 'c']
    >>> cq = pdp.cq.AllColumns(subset=True)
    >>> cq(df)
    ['a', 'b']
    >>> cq(df2)
    ['b']
    """

    class _SelectAllColumns(object):

        def __call__(self, X):
            return list(X.columns)

    def __init__(self, **kwargs):
        kwargs['func'] = AllColumns._SelectAllColumns()
        super().__init__(**kwargs)

    def __repr__(self):
        return "<ColumnQualifier: Qualify all columns>"


class ByColumnCondition(ColumnQualifier):
    """A fittable column qualifier based on a per-column condition.

    Parameters
    ----------
    cond : callable
        A callable that given an input pandas.Series object returns a boolean
        value.
    safe : bool, default False
        If set to True, every call to given condition `cond` is is wrapped in
        a way that interprets every raised exception as a returned False value.
        This is useful when generating qualifiers based on conditions that
        assume a specific datatype for the checked column.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of `ColumnQualifier` for
        details.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame(
    ...    [[1, 2, 'A'],[4, 1, 'C']], [1,2], ['age', 'count', 'grade'])
    >>> cq = pdp.cq.ByColumnCondition(lambda s: s.sum() > 3, safe=True)
    >>> cq(df)
    ['age']
    """

    class _SafeCond(object):

        def __init__(self, cond):
            self.cond = cond

        def __call__(self, series):
            try:
                return self.cond(series)
            except Exception:
                return False

    class _ColumnConditionChecker(object):

        def __init__(self, cond):
            self.cond = cond

        def __call__(self, X):
            return list([
                lbl for lbl, series in X.iteritems()
                if self.cond(series)
            ])

    def __init__(self, cond, safe=False, **kwargs):
        self._cond = cond
        if safe:
            self._cond = ByColumnCondition._SafeCond(cond)
        kwargs['func'] = ByColumnCondition._ColumnConditionChecker(self._cond)
        super().__init__(**kwargs)


class ByLabels(ColumnQualifier):
    """Selects all columns with the given label or labels.

    Parameters
    ----------
    labels : single label or list-like
        Column labels which qualify.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of `ColumnQualifier` for
        details.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame(
    ...    [[8,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
    >>> cq = pdp.cq.ByLabels('num')
    >>> cq(df)
    ['num']
    >>> cq = pdp.cq.ByLabels(['chr', 'nur'])
    >>> cq(df)
    ['chr', 'nur']
    >>> cq = pdp.cq.ByLabels(['num', 'foo'])
    >>> cq(df)
    ['num']
    """

    class _LabelsQualifierFunc(object):

        def __init__(self, labels):
            self.labels = labels

        def __call__(self, X):
            return [
                lbl for lbl in X.columns
                if lbl in self.labels
            ]

    def __init__(self, labels, **kwargs):
        if isinstance(labels, str) or not hasattr(labels, '__iter__'):
            labels = [labels]
        self._labels = labels
        self._labels_str = _list_str(self._labels)
        cqfunc = ByLabels._LabelsQualifierFunc(self._labels)
        cqfunc.__doc__ = f"Columns with labels in {self._labels_str}"
        self.__doc__ = cqfunc.__doc__
        kwargs['func'] = cqfunc
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<ColumnQualifier: By labels in {self._labels_str}>"


def columns_to_qualifier(columns):
    """Converts the given columns parameter to an equivalent column qualifier.

    Parameters
    ----------
    columns : single label, list-like or callable
        The label, or an iterable of labels, of columns. Alternatively,
        this parameter can be assigned a callable returning an iterable of
        labels from an input pandas.DataFrame. See `pdpipe.cq`.

    Returns
    -------
    qualifier : ColumnQualifier
        The equivalent ColumnQualifier object.

    Example
    -------
    >>> import pdpipe as pdp;
    >>> pdp.cq.columns_to_qualifier('nu')
    <ColumnQualifier: By labels in nu>
    >>> pdp.cq.columns_to_qualifier(['nu', 'bu'])
    <ColumnQualifier: By labels in nu, bu>
    >>> pdp.cq.columns_to_qualifier(lambda df: [l for l in df.columns])
    <ColumnQualifier: Qualify columns by function>
    """
    if callable(columns):
        if isinstance(columns, ColumnQualifier):
            return columns
        return ColumnQualifier(columns, fittable=False)
    return ByLabels(columns)


class StartWith(ColumnQualifier):
    """Selectes all columns that start with the given string.

    Parameters
    ----------
    prefix : str
        The prefix which qualifies columns.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of `ColumnQualifier` for
        details.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame(
    ...    [[8,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
    >>> cq = pdp.cq.StartWith('nu')
    >>> cq
    <ColumnQualifier: Columns starting with nu>
    >>> cq(df)
    ['num', 'nur']
    """

    @staticmethod
    def _safe_startwith(string, prefix):
        try:
            return string.startswith(prefix)
        except AttributeError:
            return False

    class _StartWithFunc(object):

        def __init__(self, prefix):
            self.prefix = prefix

        def __call__(self, X):
            return [
                lbl for lbl in X.columns
                if StartWith._safe_startwith(lbl, self.prefix)
            ]

    def __init__(self, prefix, **kwargs):
        self._prefix = prefix
        cqfunc = StartWith._StartWithFunc(prefix)
        cqfunc.__doc__ = f"Columns that start with {self._prefix}"
        self.__doc__ = cqfunc.__doc__
        kwargs['func'] = cqfunc
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<ColumnQualifier: Columns starting with {self._prefix}>"


class OfDtypes(ColumnQualifier):
    """Selectes all columns that are of a given dtypes.

    Use `dtypes=np.number` to qualify all numeric columns.

    Parameters
    ----------
    dtypes : object or list of objects
        The dtype or dtypes which qualify columns. Support all valid arguments
        to the `include` parameter of pandas.DataFrame.select_dtypes().
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of `ColumnQualifier` for
        details.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp; import numpy as np;
    >>> df = pd.DataFrame(
    ...    [[8.2,'a',5],[5.1,'b',7]], [1,2], ['ph', 'grade', 'age'])
    >>> cq = pdp.cq.OfDtypes(np.number)
    >>> cq(df)
    ['ph', 'age']
    >>> cq = pdp.cq.OfDtypes([np.number, object])
    >>> cq(df)
    ['ph', 'grade', 'age']
    >>> cq = pdp.cq.OfDtypes(np.int64)
    >>> cq
    <ColumnQualifier: With dtypes in <class 'numpy.int64'>>
    >>> cq(df)
    ['age']
    """

    class _OfDtypeFunc(object):

        def __init__(self, dtypes):
            self.dtypes = dtypes

        def __call__(self, X):
            return list(X.select_dtypes(include=self.dtypes).columns)

    def __init__(self, dtypes, **kwargs):
        self._dtypes = dtypes
        self._dtypes_str = _list_str(self._dtypes)
        cqfunc = OfDtypes._OfDtypeFunc(dtypes)
        cqfunc.__doc__ = f"Columns of dtypes {self._dtypes_str}"
        self.__doc__ = cqfunc.__doc__
        kwargs['func'] = cqfunc
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<ColumnQualifier: With dtypes in {self._dtypes_str}>"


class OfNumericDtypes(OfDtypes):
    """Selectes all columns that are of a numeric dtypes.

    Parameters
    ----------
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of `ColumnQualifier` for
        details.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp; import numpy as np;
    >>> df = pd.DataFrame(
    ...    [[8.2,'a',5],[5.1,'b',7]], [1,2], ['ph', 'grade', 'age'])
    >>> cq = pdp.cq.OfNumericDtypes()
    >>> cq
    <ColumnQualifier: With dtypes in <class 'numpy.number'>>
    >>> cq(df)
    ['ph', 'age']
    """

    def __init__(self, **kwargs):
        kwargs['dtypes'] = np.number
        super().__init__(**kwargs)


class WithAtMostMissingValues(ColumnQualifier):
    """Selectes all columns with no more than X missing values.

    Parameters
    ----------
    n_missing : int
        The maximum number of missing values with which columns can still
        qualify.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of `ColumnQualifier` for
        details.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp; import numpy as np;
    >>> df = pd.DataFrame(
    ...    [[None, 1, 2],[None, None, 5]], [1,2], ['ph', 'grade', 'age'])
    >>> cq = pdp.cq.WithAtMostMissingValues(1)
    >>> cq
    <ColumnQualifier: With at most 1 missing values>
    >>> cq(df)
    ['grade', 'age']
    """

    class _AtMostFunc(object):

        def __init__(self, n_missing):
            self._n_missing = n_missing

        def __call__(self, X):
            return list(X.columns[X.isna().sum() <= self._n_missing])

    def __init__(self, n_missing, **kwargs):
        self._n_missing = n_missing
        cqfunc = WithAtMostMissingValues._AtMostFunc(n_missing)
        cqfunc.__doc__ = (
            f"Columns with at most {self._n_missing} missing values"
        )
        self.__doc__ = cqfunc.__doc__
        kwargs['func'] = cqfunc
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<ColumnQualifier: " \
               f"With at most {self._n_missing} missing values>"


class WithoutMissingValues(WithAtMostMissingValues):
    """Selectes all columns with no missing values.

    Parameters
    ----------
    **kwargs
        Accepts all keyword arguments of the constructor of ColumnQualifier.
        See the documentation of `ColumnQualifier` for details.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp; import numpy as np;
    >>> df = pd.DataFrame(
    ...    [[None, 1, 2],[None, None, 5]], [1,2], ['ph', 'grade', 'age'])
    >>> cq = pdp.cq.WithoutMissingValues()
    >>> cq
    <ColumnQualifier: Without missing values>
    >>> cq(df)
    ['age']
    """

    def __init__(self, **kwargs):
        kwargs['n_missing'] = 0
        super().__init__(**kwargs)

    def __repr__(self):
        return "<ColumnQualifier: Without missing values>"



"""Configuring pytest for pdpipe tests."""

import os
import sys
from pathlib import Path

import xdg
import pytest


# make
sys.path.append(os.path.join(os.path.dirname(__file__), 'helpers'))


@pytest.fixture(scope='session', autouse=False)
def pdpipe_tests_dir_path() -> str:
    cache_dpath: Path
    cache_dpath = xdg.xdg_cache_home()
    tests_dpath = cache_dpath / 'pdpipe' / 'tests'
    tests_dpath = str(tests_dpath)
    os.makedirs(tests_dpath, exist_ok=True)
    return tests_dpath



"""Configuration control for pdpipe."""

import birch


class CfgKey():
    LOAD_STAGE_ATTRIBUTES = 'LOAD_STAGE_ATTRIBUTES'


CFG = birch.Birch(
    namespace='pdpipe',
    defaults={
        CfgKey.LOAD_STAGE_ATTRIBUTES: 'True'
    },
    default_casters={
        CfgKey.LOAD_STAGE_ATTRIBUTES: birch.casters.true_false_caster,
    },
)

LOAD_STAGE_ATTRIBUTES = CFG[CfgKey.LOAD_STAGE_ATTRIBUTES]



"""Testing basic pipeline stages."""

import pytest
import pandas as pd

from pdpipe.basic_stages import ConditionValidator
from pdpipe.cond import HasNoMissingValues, HasNoColumn
from pdpipe.exceptions import FailedConditionError


DF1 = pd.DataFrame([[1, 4], [4, None], [1, 11]], [1, 2, 3], ['a', 'b'])


def test_condition_validator_basic():
    stage = ConditionValidator(HasNoMissingValues())
    with pytest.raises(FailedConditionError):
        stage(DF1)


def test_condition_validator_ignore():
    stage = ConditionValidator(HasNoMissingValues(), errors='ignore')
    res_df = stage(DF1)
    assert 1 in res_df.index
    assert 2 in res_df.index
    assert 3 in res_df.index


def test_condition_validator_multi_default():
    stage = ConditionValidator([HasNoMissingValues(), HasNoColumn('k')])
    with pytest.raises(FailedConditionError):
        stage(DF1)


def test_condition_validator_multi_any():
    stage = ConditionValidator(
        conditions=[HasNoMissingValues(), HasNoColumn('k')],
        reducer=any,
    )
    res_df = stage(DF1, verbose=True)
    assert 1 in res_df.index
    assert 2 in res_df.index
    assert 3 in res_df.index


def test_condition_validator_bad_input():
    stage = ConditionValidator([5, HasNoColumn('k')])
    with pytest.raises(ValueError):
        stage(DF1)


def test_condition_validator_lambda_pass():
    stage = ConditionValidator(
        conditions=[lambda df: True, HasNoColumn('k')],
    )
    res_df = stage(DF1, verbose=True)
    assert 1 in res_df.index
    assert 2 in res_df.index
    assert 3 in res_df.index


def test_condition_validator_lambda_fail():
    stage = ConditionValidator([lambda df: False, HasNoColumn('k')])
    with pytest.raises(FailedConditionError):
        stage(DF1, verbose=True)

    # Used as documentation example
    with pytest.raises(FailedConditionError):
        ConditionValidator(lambda df: len(df.columns) == 5).apply(DF1)


def test_condition_validator_custom_func():
    def _foo(df: pd.DataFrame) -> bool:
        return False
    _foo.__doc__ = None
    stage = ConditionValidator([_foo, HasNoColumn('k')])
    with pytest.raises(FailedConditionError):
        stage(DF1, verbose=True)



"""Custom types for pdpipe."""

from typing import Union, List, Callable

import numpy
import pandas


ColumnsParamType = Union[object, List[object], Callable]
ColumnLabelsType = Union[object, List[object]]

_series_operand_types = [
    pandas.Series,
    int,
    float,
    complex,
    int,
    bool,
    bytes,
    str,
    memoryview,
    numpy.int8,
    numpy.uint8,
    numpy.float16,
    numpy.timedelta64,
    numpy.object_,
    numpy.int16,
    numpy.uint16,
    numpy.float32,
    numpy.complex64,
    numpy.bytes_,
    numpy.int32,
    numpy.uint32,
    numpy.float64,
    numpy.str_,
    numpy.int64,
    numpy.uint64,
    numpy.bool_,
    numpy.void,
    numpy.longlong,
    numpy.ulonglong,
    numpy.datetime64,
]

try:
    _series_operand_types.extend([
        numpy.complex128,
        numpy.float128,
    ])
except AttributeError:  # pragma: no cover
    # were on a system with no 128-bit floats
    pass
try:
    _series_operand_types.append(numpy.complex256)
except AttributeError:  # pragma: no cover
    pass


SeriesOperandTypesTuple = tuple(_series_operand_types)



"""pytest file built from tests/cond/cond_example.md"""


def session_00001_line_2():
    r"""
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame(
    ...    [[8,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
    >>> cond = ~ pdp.cond.HasAllColumns(['num', 'chr'])
    >>> cond(df)
    False
    >>> cond = ~ pdp.cond.HasAllColumns(['num','go'])
    >>> cond(df)
    True
    """


def session_00002_line_14():
    r"""
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[8, None],[5, 2]], [1,2], ['foo', 'bar'])
    >>> col_cond = pdp.cond.HasAllColumns(['foo', 'bar'])
    >>> missing_cond = pdp.cond.HasNoMissingValues()
    >>> (col_cond | missing_cond)(df)
    True
    >>> (col_cond & missing_cond)(df)
    False
    >>> df = pd.DataFrame([[8, 9],[5, 2]], [1,2], ['foo', 'bar'])
    >>> (col_cond & missing_cond)(df)
    True
    """


def session_00003_line_28():
    r"""
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[8, None],[5, 2]], [1,2], ['foo', 'bar'])
    >>> col_cond = pdp.cond.HasAllColumns(['foo', 'bar'])
    >>> missing_cond = pdp.cond.HasNoMissingValues()
    >>> (col_cond ^ missing_cond)(df)
    True
    >>> df = pd.DataFrame([[8, 9],[5, 2]], [1,2], ['foo', 'bar'])
    >>> (col_cond ^ missing_cond)(df)
    False
    """



"""Testing basic pipeline stages."""

import pytest
import pandas as pd
# from numpy.testing import assert_approx_equal

from pdpipe.sklearn_stages import Decompose
from pdpipe.exceptions import PipelineApplicationError

from sklearn.decomposition import PCA


def _some_df1():
    return pd.DataFrame(
        data=[[3.2, 0.3, "A"], [7.2, 0.35, "B"], [12.1, 0.29, "C"]],
        index=[1, 2, 3],
        columns=["ph", "gt", "lbl"],
    )


def _some_df1b():
    return pd.DataFrame(
        data=[[3.8, 0.45, "A"], [7.7, 0.31, "B"], [11.15, 0.33, "C"]],
        index=[1, 2, 3],
        columns=["ph", "gt", "lbl"],
    )


def _bad_dtype_df1():
    return pd.DataFrame(
        data=[[3.2, "H", "A"], [7.2, "I", "B"], [12.1, "P", "C"]],
        index=[1, 2, 3],
        columns=["ph", "gt", "lbl"],
    )


def _some_df2():
    return pd.DataFrame(
        data=[[3.2, 0.3], [7.2, 0.35], [12.1, 0.29]],
        index=[1, 2, 3],
        columns=["ph", "gt"],
    )


def _some_df2b():
    return pd.DataFrame(
        data=[[3.8, 0.45], [7.7, 0.31], [11.15, 0.33]],
        index=[1, 2, 3],
        columns=["ph", "gt"],
    )


def test_decompose():
    df = _some_df2()
    decomp_stage = Decompose(PCA(), n_components=2)
    res_df = decomp_stage(df)
    assert "ph" not in res_df.columns
    assert "gt" not in res_df.columns
    assert "mdc0" in res_df.columns
    assert "mdc1" in res_df.columns

    # see only transform (no fit) when already fitted
    df2 = _some_df2b()
    res_df2 = decomp_stage(df2)
    assert "ph" not in res_df2.columns
    assert "gt" not in res_df2.columns
    assert "mdc0" in res_df2.columns
    assert "mdc1" in res_df2.columns

    # check fit_transform when already fitted
    df3 = _some_df2b()
    res_df3 = decomp_stage.fit_transform(df3)
    assert "ph" not in res_df3.columns
    assert "gt" not in res_df3.columns
    assert "mdc0" in res_df3.columns
    assert "mdc1" in res_df3.columns


def test_decompose_with_exclude_cols():
    df = _some_df1()
    decomp_stage = Decompose(PCA(), n_components=2, exclude_columns=["lbl"])
    res_df = decomp_stage(df)
    assert list(res_df.columns) == ["lbl", "mdc0", "mdc1"]
    assert "ph" not in res_df.columns
    assert "gt" not in res_df.columns

    # see only transform (no fit) when already fitted
    df2 = _some_df1b()
    res_df2 = decomp_stage(df2)
    assert "lbl" in res_df2.columns
    assert "ph" not in res_df2.columns
    assert "gt" not in res_df2.columns
    assert "mdc0" in res_df2.columns
    assert "mdc1" in res_df2.columns


def test_decompose_without_drop():
    """Basic binning test."""
    df = _some_df1()
    decomp_stage = Decompose(PCA(), n_components=2, drop=False)
    res_df = decomp_stage(df)
    assert list(res_df.columns) == ["ph", "gt", "lbl", "mdc0", "mdc1"]

    # see only transform (no fit) when already fitted
    df2 = _some_df1b()
    res_df2 = decomp_stage(df2)
    assert list(res_df2.columns) == ["ph", "gt", "lbl", "mdc0", "mdc1"]


def test_decompose_fit_transform_exception():
    df1 = _some_df1()
    decomp_stage = Decompose(PCA(), n_components=2, columns=['ph', 'lbl'])
    with pytest.raises(PipelineApplicationError):
        decomp_stage(df1)


def test_decompose_transform_exception():
    df1 = _some_df1()
    decomp_stage = Decompose(PCA(), n_components=2, exmsg="ERR")
    decomp_stage(df1)

    # test transform exception
    df2 = _bad_dtype_df1()
    with pytest.raises(PipelineApplicationError):
        decomp_stage(df2)


def test_decompose_with_lbl_format():
    df = _some_df2()
    decomp_stage = Decompose(PCA(), lbl_format='pca{}', n_components=2)
    res_df = decomp_stage(df)
    assert "ph" not in res_df.columns
    assert "gt" not in res_df.columns
    assert "pca0" in res_df.columns
    assert "pca1" in res_df.columns

    # see only transform (no fit) when already fitted
    df2 = _some_df2b()
    res_df2 = decomp_stage(df2)
    assert "ph" not in res_df2.columns
    assert "gt" not in res_df2.columns
    assert "pca0" in res_df2.columns
    assert "pca1" in res_df2.columns



"""PdPipeline stages that transform the optional label column."""

from pdpipe.core import PdPipelineStage

from .exceptions import (
    PipelineInitializationError,
    UnexpectedPipelineMethodCallError,
)


class DropLabelsByValues(PdPipelineStage):
    """A pipeline stage that drop values from the input label series.

    Parameters
    ----------
    in_set : iterable of object, optional
        Drop all labels in the input set of values.
    in_ranges : iterable of iterables of scalars, optional
        Drop all labels in the input ranges of values.
    not_in_set : iterable of object, optional
        Drop all labels not in the input set of values.
    not_in_ranges : iterable of iterables of scalars, optional
        Drop all labels not in the input ranges of values.


    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> data = [[3.2, 31], [7.2, 33], [12.1, 28]]
    >>> X = pd.DataFrame(data, [1,2,3], ["ph","temp"])
    >>> y = pd.Series(["acd", "alk", "alk"])
    >>> drop_labels = pdp.DropLabelsByValues(in_set=["acd"])
    >>> X, y = drop_labels(X, y)
    >>> y
    2    alk
    3    alk
    dtype: object
    >>> X
         ph  temp
    2   7.2    33
    3  12.1    28
    """

    def __init__(
        self,
        in_set=None,
        in_ranges=None,
        not_in_set=None,
        not_in_ranges=None,
        **kwargs: object,
    ):
        self.in_set = in_set
        self.in_ranges = in_ranges
        self.not_in_set = not_in_set
        self.not_in_ranges = not_in_ranges
        super_kwargs = {
            'desc': "Encode label values",
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X, y):
        return y is not None

    def _transform(self, X, verbose):
        raise UnexpectedPipelineMethodCallError(
            "EncodeLabel._transform() is not expected to be called!")

    def _transform_Xy(self, X, y, verbose):
        post_y = y
        if self.in_set is not None:
            post_y = post_y.loc[~ post_y.isin(self.in_set)]
        elif self.in_ranges is not None:
            to_drop = y.copy()
            to_drop.loc[:] = False
            for in_range in self.in_ranges:
                to_drop = to_drop | (y.between(*in_range))
            post_y = post_y.loc[~to_drop]
        elif self.not_in_set is not None:
            post_y = y.isin(self.not_in_set)
        elif self.not_in_ranges is not None:
            to_keep = y.copy()
            to_keep.loc[:] = False
            for in_range in self.not_in_ranges:
                to_keep = to_keep | (y.between(*in_range))
            post_y = post_y.loc[to_keep]
        else:
            raise PipelineInitializationError(
                "DropLabelsByValues: No drop conditions specified.")
        return X, post_y



"""Defines pipelines for processing pandas.DataFrame-based datasets."""

import re
import sys
import abc
import time
import inspect
import collections
import textwrap
from typing import Tuple, Union, Iterable, Optional

import numpy
import pandas

try:
    from pympler.asizeof import asizeof
except ImportError:
    from sys import getsizeof as asizeof

from .cfg import (
    LOAD_STAGE_ATTRIBUTES,
)
from .cq import is_fittable_column_qualifier, AllColumns
from .shared import (
    _get_args_list,
    _always_true,
)
from .exceptions import (
    FailedPreconditionError,
    FailedPostconditionError,
    UnfittedPipelineStageError,
    PipelineApplicationError
)


# === loading stage attributes ===

def __get_append_stage_attr_doc(class_obj: object) -> str:
    doc = class_obj.__doc__
    if doc is None:  # pragma: no cover
        return
    first_line = doc[0:doc.find('.') + 1]
    if "An" in first_line:
        new_first_line = first_line.replace("An", "Creates and adds an", 1)
    else:
        new_first_line = first_line.replace("A", "Creates and adds a", 1)
    new_first_line = new_first_line[0:-1] + (
        " to this pipeline stage.")
    return doc.replace(first_line, new_first_line, 1)


def __load_stage_attribute__(class_obj: object) -> None:

    def _append_stage_func(self, *args, **kwds):
        # self is always a PdPipelineStage
        return self + class_obj(*args, **kwds)
    _append_stage_func.__doc__ = __get_append_stage_attr_doc(class_obj)
    _append_stage_func.__name__ = class_obj.__name__  # .lower()
    _append_stage_func.__signature__ = inspect.signature(class_obj.__init__)
    setattr(PdPipelineStage, class_obj.__name__, _append_stage_func)

    # unbound_method = types.MethodType(_append_stage_func, class_obj)
    # setattr(class_obj, class_obj.__name__, unbound_method)


def __load_stage_attributes_from_module__(module_name: object) -> None:
    if not LOAD_STAGE_ATTRIBUTES:
        return  # pragma: no cover
    module_obj = sys.modules[module_name]
    for name, obj in inspect.getmembers(module_obj):
        if inspect.isclass(obj) and obj.__module__ == module_name:
            class_obj = getattr(module_obj, name)
            if issubclass(class_obj, PdPipelineStage) and (
                    class_obj.__name__ != 'PdPipelineStage'):
                __load_stage_attribute__(class_obj)


# === basic classes ===

class PdpApplicationContext():
    """An object encapsulating the application context of a pipeline.

    It is meant to communicate data, information and variables between
    different stages of a pipeline.

    Parameters
    ----------
    fit_context : PdpApplicationContext, optional
        Another application context object, representing the application
        context of a previous fit of the pipelline this application context
        is initialized for. Optional.
    """

    def __init__(
        self,
        fit_context: Optional['PdpApplicationContext'] = None,
    ) -> None:
        self._locked = False
        self._fit_context = fit_context
        self._dict = {}

    def __getitem__(self, key: object) -> object:
        return self._dict[key]

    def __setitem__(self, key: object, value: object) -> None:
        if not self._locked:
            self._dict[key] = value

    def __delitem__(self, key: object) -> object:
        if not self._locked:
            self._dict.__delitem__(key)

    def get(self, key: object, default: object = None) -> object:
        """Return the value for key if key is in the dictionary, else default
        If default is not given, it defaults to None, so that this method never
        raises a KeyError."""
        return self._dict.get(key, default)

    def items(self):
        """Return a new view of the context’s items ((key, value) pairs)."""
        return self._dict.items()

    def keys(self):
        """Return a new view of the context's keys."""
        return self._dict.keys()

    def pop(self, key: object, default: object) -> object:
        """If key is in the dictionary, remove it and return its value, else
        return default. If default is not given and key is not in the
        dictionary, a KeyError is raised.
        """
        if not self._locked:
            return self._dict.pop(key, default)
        return self._dict.__getitem__(key)

    def clear(self) -> None:
        """Remove all items from the dictionary."""
        if not self._locked:
            self._dict.clear()

    def popitem(self) -> object:
        """Not implemented!"""
        raise NotImplementedError

    def update(self, other: dict) -> None:
        """Update the dictionary with the key/value pairs from other,
        overwriting existing keys. Return None.
        update() accepts either another dictionary object or an iterable of
        key/value pairs (as tuples or other iterables of length two). If
        keyword arguments are specified, the dictionary is then updated with
        those key/value pairs: d.update(red=1, blue=2).
        """
        if not self._locked:
            self._dict.update(other)

    def lock(self) -> None:
        """Locks this application context for changes."""
        self._locked = True

    def fit_context(self) -> 'PdpApplicationContext':
        """Returns a locked PdpApplicationContext object of a previous fit."""
        return self._fit_context


class PdPipelineStage(abc.ABC):
    """A stage of a pandas DataFrame-processing pipeline.

    Parameters
    ----------
    exraise : bool, default True
        If true, a pdpipe.FailedPreconditionError is raised when this
        stage is applied to a dataframe for which the precondition does
        not hold. Otherwise the stage is skipped. Additionally, if true, a
        pdpipe.FailedPostconditionError is raised if an expected post-condition
        does not hold for an output dataframe (after pipeline application).
        Otherwise pipeline application continues uninterrupted.
    exmsg : str, default None
        The message of the exception that is raised on a failed
        precondition if exraise is set to True. A default message is used
        if None is given.
    desc : str, default None
        A short description of this stage, used as its string representation.
        A default description is used if None is given.
    prec : callable, default None
        This can be assigned a callable that returns boolean values for input
        dataframes, which will be used to determine whether input dataframes
        satisfy the preconditions for this pipeline stage (see the `exraise`
        parameter for the behaviour of failed preconditions). See `pdpipe.cond`
        for more information on specialised Condition objects.
    post : callable, default None
        This can be assigned a callable that returns boolean values for input
        dataframes, which will be used to determine whether input dataframes
        satisfy the postconditions for this pipeline stage (see the `exraise`
        parameter for the behaviour of failed postconditions). See
        `pdpipe.cond` for more information on specialised Condition objects.
    skip : callable, default None
        This can be assigned a callable that returns boolean values for input
        dataframes, which will be used to determine whether this stage should
        be skipped for input dataframes - if the callable returns True for an
        input dataframe, this stage will be skipped. See `pdpipe.cond` for more
        information on specialised Condition objects.
    name : str, default ''
        The name of this stage. Pipelines can be sliced by this name.

    Attributes
    ----------
    fit_context : `PdpApplicationContext`
        An application context object that is only re-initialized before
        `fit_transform` calls, and is locked after pipeline application. It is
        injected into the PipelineStage by the encapsulating pipeline object.
    application_context : `PdpApplicationContext`
        An application context object that is re-initialized before every
        pipeline application (so, also during transform operations of fitted
        pipelines), and is locked after pipeline application.It is injected
        into the PipelineStage by the encapsulating pipeline object.
    """

    _DEF_EXC_MSG = 'Precondition failed in stage {}!'
    _DEF_DESCRIPTION = 'A pipeline stage.'
    _INIT_KWARGS = ['exraise', 'exmsg', 'desc', 'prec', 'skip', 'name']

    def __init__(
        self,
        exraise: Optional[bool] = True,
        exmsg: Optional[str] = None,
        desc: Optional[str] = None,
        prec: Optional[callable] = None,
        post: Optional[callable] = None,
        skip: Optional[callable] = None,
        name: Optional[str] = '',
    ) -> None:
        if not isinstance(name, str):
            raise ValueError(
                f"'name' must be a str, not {type(name).__name__}."
            )
        if desc is None:
            desc = PdPipelineStage._DEF_DESCRIPTION
        if exmsg is None:
            exmsg = PdPipelineStage._DEF_EXC_MSG.format(desc)

        # save input parameters
        self._exraise = exraise
        self._exmsg = exmsg
        self._exmsg_post = exmsg.replace(
            'precondition', 'postcondition').replace(
            'Precondition', 'Postcondition')
        self._desc = desc
        self._prec_arg = prec
        self._post_arg = post
        self._skip = skip
        self._appmsg = f"{name + ': ' if name else ''}{desc}"
        self._name = name

        # inner stuff initializations
        self._is_an_Xy_transformer = False
        self._is_an_Xy_fit_transformer = False
        if hasattr(self.__class__, '_transform_Xy'):
            self._is_an_Xy_transformer = True
        if hasattr(self.__class__, '_fit_transform_Xy'):
            self._is_an_Xy_fit_transformer = True
        self.fit_context: PdpApplicationContext = None
        self.application_context: PdpApplicationContext = None
        self.is_fitted = False

    @classmethod
    def _init_kwargs(cls):
        return cls._INIT_KWARGS

    @classmethod
    def _split_kwargs(cls, kwargs: dict) -> dict:
        """Splits the given kwargs dict into init and non-init kwargs.

        Parameters
        ----------
        kwargs : dict
            The kwargs dict to split.

        Returns
        -------
        init_kwargs : dict
            The init kwargs dict.
        other_kwargs : dict
            The non-init kwargs dict.
        """
        init_kwargs = {
            k: v for k, v in kwargs.items() if k in cls._INIT_KWARGS}
        other_kwargs = {
            k: v for k, v in kwargs.items() if k not in cls._INIT_KWARGS}
        return init_kwargs, other_kwargs

    _POS_ARG_MISMTCH_PAT = re.compile(
        r'\d positional argument[s]? but \d (were|was) given')

    @abc.abstractmethod
    def _prec(
        self,
        X: pandas.DataFrame,
        y: Optional[pandas.Series] = None,
    ) -> bool:  # pylint: disable=R0201,W0613
        """Returns True if this stage can be applied to the given dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The input dataframe.
        y : pandas.Series, optional
            A possible label column for processing of supervised learning
            datasets. Might also be inspected, if provided.

        Returns
        -------
        bool
            True if this stage this stage can be applied to the given
            dataframe; False otherwise.
        """
        raise NotImplementedError

    def _compound_prec(
        self,
        X: pandas.DataFrame,
        y: Optional[pandas.Series] = None,
    ) -> bool:
        if self._prec_arg:
            try:
                return self._prec_arg(X, y)
            except TypeError as e:
                if len(PdPipelineStage._POS_ARG_MISMTCH_PAT.findall(
                        str(e))) > 0:
                    return self._prec_arg(X)
                raise e  # pragma: no cover
        # now, do the same for the _prec abstractmethod, so as to support
        # implementations only including X in their signatur
        if y is None:
            return self._prec(X)
        try:
            return self._prec(X, y)
        except TypeError as e:
            if len(PdPipelineStage._POS_ARG_MISMTCH_PAT.findall(str(e))) > 0:
                return self._prec(X)
            raise e  # pragma: no cover

    def _post(
        self,
        X: pandas.DataFrame,
        y: Optional[pandas.Series] = None,
    ) -> bool:  # pylint: disable=R0201,W0613
        """Returns True if this stage resulted in an expected output frame."""
        return True

    def _compound_post(
        self,
        X: pandas.DataFrame,
        y: Optional[pandas.Series] = None,
    ) -> bool:
        """An inner implementation of the post-condition functionality.

        Uses the constructor provided post-condition, if one was provided.
        Otherwise, uses the build-it function.
        """
        if self._post_arg:
            try:
                return self._post_arg(X, y)
            except TypeError as e:
                if len(PdPipelineStage._POS_ARG_MISMTCH_PAT.findall(
                        str(e))) > 0:
                    return self._post_arg(X)
                raise e  # pragma: no cover
        if y is None:
            return self._post(X)
        try:
            return self._post(X, y)
        except TypeError as e:
            if len(PdPipelineStage._POS_ARG_MISMTCH_PAT.findall(str(e))) > 0:
                return self._post(X)
            raise e  # pragma: no cover

    def _fit_transform(
        self,
        X: pandas.DataFrame,
        verbose: bool = False,
    ) -> pandas.DataFrame:
        """Fits this stage and transforms the input dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The input dataframe.
        verbose : bool, default False
            If True, might print informative messages.

        Returns
        -------
        pandas.DataFrame
            The transformed dataframe.
        """
        return self._transform(X, verbose=verbose)

    def _is_fittable(self) -> bool:
        if self.__class__._fit_transform == PdPipelineStage._fit_transform:
            return False
        return True

    def _raise_precondition_error(self) -> None:
        try:
            raise FailedPreconditionError(
                f"{self._exmsg} [Reason] {self._prec_arg.error_message}")
        except AttributeError:
            raise FailedPreconditionError(self._exmsg)

    def _raise_postcondition_error(self) -> None:
        try:
            raise FailedPostconditionError(
                f"{self._exmsg_post} [Reason] {self._post_arg.error_message}")
        except AttributeError:
            raise FailedPostconditionError(self._exmsg_post)

    @abc.abstractmethod
    def _transform(
        self,
        X: pandas.DataFrame,
        verbose: bool = False,
    ) -> pandas.DataFrame:
        """Transforms an input dataframe without fitting this stage.

        Parameters
        ----------
        X : pandas.DataFrame
            The input dataframe.
        verbose : bool, optional
            If True, prints the progress of the transformation.

        Returns
        -------
        pandas.DataFrame
            The transformed dataframe.
        """
        raise NotImplementedError("_transform method not implemented!")

    @staticmethod
    def _cast_y_to_series(
        X: pandas.DataFrame,
        y: Union[pandas.Series, numpy.array, Iterable[object]],
    ) -> pandas.Series:
        """Cast the y labels input to a correclty-indexed pandas.Series.


        Parameters
        ----------
        X : pandas.DataFrame
            The input dataframe.
        y : pandas.Series, or numpy.ndarray, or Iterable[object]
            The label series/array.

        Returns
        -------
        pandas.Series
            The label array as a pandas.Series object.
        """
        if len(X) != len(y):
            raise ValueError("X and y must have the same length!")
        if isinstance(y, pandas.Series):
            post_y = pandas.Series(data=y.values, index=X.index)
        else:
            post_y = pandas.Series(data=y, index=X.index)
        return post_y

    @staticmethod
    def _align_Xy(
        X: pandas.DataFrame,
        y: pandas.Series,
    ) -> Tuple[pandas.DataFrame, pandas.Series]:
        """Aligns the input dataframe and label series.

        The input dataframe and label series are assumed to have been indexed
        by the same index before a possible transformation to one of them,
        which might have dropped some rows/values in onf of them, and/or
        reorded the rows/values in one of them.

        Parameters
        ----------
        X : pandas.DataFrame
            The input dataframe.
        y : pandas.Series
            The label series.

        Returns
        -------
        pandas.DataFrame
            The aligned dataframe.
        pandas.Series
            The aligned label series.
        """
        # if values were dropped from y, reindex X according to y
        if len(y) < len(X):
            return X.loc[y.index], y
        # otherwise, transformation was almost certainly done on X (might have
        # been reordered), so reindex y according to X
        return X, y.loc[X.index]

    def _should_skip(self, X, y) -> bool:
        if self._skip:
            try:
                return self._skip(X, y)
            except TypeError as e:
                if len(PdPipelineStage._POS_ARG_MISMTCH_PAT.findall(
                        str(e))) > 0:
                    return self._skip(X)
                else:
                    raise e  # pragma: no cover
        return False

    def apply(
        self,
        X: pandas.DataFrame,
        y: Optional[pandas.Series] = None,
        exraise: Optional[bool] = None,
        verbose: Optional[bool] = False,
    ) -> Union[pandas.DataFrame, Tuple[pandas.DataFrame, pandas.Series]]:
        """Applies this pipeline stage to the given dataframe.

        If the stage is not fitted fit_transform is called. Otherwise,
        transform is called.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to which this pipeline stage will be applied.
        y : pandas.Series, optional
            A possible label column for processing of supervised learning
            datasets. Might also be transformed, if provided.
        exraise : bool, default None
            Override preconditions and postconditions behaviour for this call.
            If None, the default behaviour of this stage is used, as determined
            by the exraise constructor parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            is checked but before the application of the pipeline stage.
            Defaults to False.

        Returns
        -------
        pandas.DataFrame or Tuple[pandas.DataFrame, pandas.Series]
            The returned dataframe. If `y` was also provided, the transformed
            `X` and `y` are returned as a tuple instead.
        """
        if exraise is None:
            exraise = self._exraise
        if self._should_skip(X, y):
            if y is not None:
                return X, y
            return X
        if y is not None:
            y = self._cast_y_to_series(X, y)
        if self._compound_prec(X, y):
            if verbose:
                msg = '- ' + '\n  '.join(textwrap.wrap(self._appmsg))
                print(msg, flush=True)
            if self.is_fitted:
                if self._is_an_Xy_transformer:
                    res_X, res_y = self._transform_Xy(
                        X, y, verbose=verbose)
                else:
                    res_X = self._transform(X, verbose=verbose)
                    res_y = y
            else:
                if self._is_an_Xy_fit_transformer:
                    res_X, res_y = self._fit_transform_Xy(
                        X, y, verbose=verbose)
                elif self._is_an_Xy_transformer:
                    res_X, res_y = self._transform_Xy(
                        X, y, verbose=verbose)
                else:
                    res_X = self._fit_transform(X, verbose=verbose)
                    res_y = y
            if exraise and not self._compound_post(X=res_X, y=res_y):
                self._raise_postcondition_error()
            # if were here, the postcondition is satisfied after a successful
            # transformation of the input dataframe, so return it
            if y is not None:
                res_X, res_y = self._align_Xy(X=res_X, y=res_y)
                return res_X, res_y
            return res_X
        # precondition doesn't hold, so if exraise==True raise an error
        if exraise:
            self._raise_precondition_error()
        # precondition doesn't hold, but we don't want to raise an error
        # as exraise is False, so return untransformed X or X, y
        if y is not None:
            return X, y
        return X

    __call__ = apply

    def fit_transform(self, X, y=None, exraise=None, verbose=False):
        """Fits this stage and transforms the given dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to transform and fit this pipeline stage by.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Override preconditions and postconditions behaviour for this call.
            If None, the default behaviour of this stage is used, as determined
            by the exraise constructor parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            is checked but before the application of the pipeline stage.
            Defaults to False.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        if exraise is None:
            exraise = self._exraise
        if self._should_skip(X, y):
            if y is not None:
                return X, y
            return X
        if y is not None:
            y = self._cast_y_to_series(X, y)
        if self._compound_prec(X, y):
            if verbose:
                msg = '- ' + '\n  '.join(textwrap.wrap(self._appmsg))
                print(msg, flush=True)
            if self._is_an_Xy_fit_transformer:
                res_X, res_y = self._fit_transform_Xy(
                    X, y, verbose=verbose)
            elif self._is_an_Xy_transformer:
                res_X, res_y = self._transform_Xy(
                    X, y, verbose=verbose)
            else:
                res_X = self._fit_transform(X, verbose=verbose)
                res_y = y
            if exraise and not self._compound_post(X=res_X, y=res_y):
                self._raise_postcondition_error()
            if y is not None:
                res_X, res_y = self._align_Xy(X=res_X, y=res_y)
                return res_X, res_y
            return res_X
        if exraise:
            self._raise_precondition_error()
        if y is not None:
            return X, y
        return X

    def fit(self, X, y=None, exraise=None, verbose=False):
        """Fits this stage without transforming the given dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to be transformed.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Override preconditions and postconditions behaviour for this call.
            If None, the default behaviour of this stage is used, as determined
            by the exraise constructor parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            is checked but before the application of the pipeline stage.
            Defaults to False.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        if exraise is None:
            exraise = self._exraise
        if self._should_skip(X, y):
            if y is not None:
                return X, y
            return X
        if y is not None:
            y = self._cast_y_to_series(X, y)
        if self._compound_prec(X, y):
            if verbose:
                msg = '- ' + '\n  '.join(textwrap.wrap(self._appmsg))
                print(msg, flush=True)
            if self._is_an_Xy_fit_transformer:
                res_X, res_y = self._fit_transform_Xy(
                    X, y, verbose=verbose)
            elif self._is_an_Xy_transformer:
                res_X, res_y = self._transform_Xy(
                    X, y, verbose=verbose)
            else:
                res_X = self._fit_transform(X, verbose=verbose)
                res_y = y
            if exraise and not self._compound_post(X=res_X, y=res_y):
                self._raise_postcondition_error()
            if y is not None:
                return X, y
            return X
        if exraise:
            self._raise_precondition_error()
        if y is not None:
            return X, y
        return X

    def transform(self, X, y=None, exraise=None, verbose=False):
        """Transforms the given dataframe without fitting this stage.

        If this stage is fittable but is not fitter, an
        UnfittedPipelineStageError is raised.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to be transformed.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Override preconditions and postconditions behaviour for this call.
            If None, the default behaviour of this stage is used, as determined
            by the exraise constructor parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            is checked but before the application of the pipeline stage.
            Defaults to False.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        if exraise is None:
            exraise = self._exraise
        if self._should_skip(X, y):
            if y is not None:
                return X, y
            return X
        if y is not None:
            y = self._cast_y_to_series(X, y)
        if self._compound_prec(X, y):
            if verbose:
                msg = '- ' + '\n  '.join(textwrap.wrap(self._appmsg))
                print(msg, flush=True)
            if self._is_fittable():
                if self.is_fitted:
                    if self._is_an_Xy_transformer:
                        res_X, res_y = self._transform_Xy(
                            X, y, verbose=verbose)
                    else:
                        res_X = self._transform(X, verbose=verbose)
                        res_y = y
                    if exraise and not self._compound_post(X=res_X, y=res_y):
                        self._raise_postcondition_error()
                    if y is not None:
                        res_X, res_y = self._align_Xy(X=res_X, y=res_y)
                        return res_X, res_y
                    return res_X
                raise UnfittedPipelineStageError(
                    "transform of an unfitted pipeline stage was called!")
            if self._is_an_Xy_transformer:
                res_X, res_y = self._transform_Xy(
                    X, y, verbose=verbose)
            else:
                res_X = self._transform(X, verbose=verbose)
                res_y = y
            if exraise and not self._compound_post(X=res_X, y=res_y):
                self._raise_postcondition_error()
            if y is not None:
                res_X, res_y = self._align_Xy(X=res_X, y=res_y)
                return res_X, res_y
            return res_X
        if exraise:
            self._raise_precondition_error()
        # precondition doesn't hold, but we don't want to raise an error
        # as exraise is False, so return untransformed X or X, y
        if y is not None:
            return X, y
        return X

    def __add__(self, other):
        if isinstance(other, PdPipeline):
            return PdPipeline([self, *other._stages])
        if isinstance(other, PdPipelineStage):
            return PdPipeline([self, other])
        return NotImplemented

    def __str__(self):
        return f"PdPipelineStage: {self._desc}"

    def __repr__(self):
        return self.__str__()

    def description(self):
        """Returns the description of this pipeline stage"""
        return self._desc

    def _mem_str(self):
        total = asizeof(self)
        lines = []
        for a in dir(self):
            if not a.startswith('__'):
                att = getattr(self, a)
                if not callable(att):
                    size = asizeof(att)
                    if size > 500000:  # pragma: no cover
                        lines.append('  - {}, {:.2f}Mb ({:0>5.2f}%)\n'.format(
                            a, size / 1000000, 100 * size / total))
                    elif size > 1000:  # pragma: no cover
                        lines.append('  - {}, {:.2f}Kb ({:0>5.2f}%)\n'.format(
                            a, size / 1000, 100 * size / total))
                    else:
                        lines.append('  - {}, {}b ({:0>5.2f}%)\n'.format(
                            a, size, 100 * size / total))
        return ''.join(lines)


class ColumnsBasedPipelineStage(PdPipelineStage):
    """A pipeline stage that operates on a subset of dataframe columns.

    Parameters
    ---------
    columns : single label, iterable or callable
        The label, or an iterable of labels, of columns to use. Alternatively,
        this parameter can be assigned a callable returning an iterable of
        labels from an input pandas.DataFrame. See `pdpipe.cq`.
    exclude_columns : single label, iterable or callable, optional
        The label, or an iterable of labels, of columns to exclude, given the
        `columns` parameter. Alternatively, this parameter can be assigned a
        callable returning a labels iterable from an input pandas.DataFrame.
        See `pdpipe.cq`. Optional. By default no columns are excluded.
    desc_temp : str, optional
        If given, assumed to be a format string, and every appearance of {} in
        it is replaced with an appropriate string representation of the columns
        parameter, and is used as the pipeline description. Ignored if `desc`
        is provided.
    none_columns : iterable, callable or str, default 'error'
        Determines how None values supplied to the 'columns' parameter should
        be handled. If set to 'error', the default, a ValueError is raised if
        None is encountered. If set to 'all', it is interpreted to mean all
        columns of input dataframes should be operated on. If an iterable is
        provided it is interpreted as the default list of columns to operate on
        when `columns=None`. If a callable is provided, it is interpreted as
        the default column qualifier that determines input columns when
        `columns=None`.
    **kwargs
        Additionally supports all constructor parameters of PdPipelineStage.
    """

    _INIT_KWARGS = [
        'columns', 'exclude_columns', 'desc_temp', 'none_columns',
    ] + PdPipelineStage._INIT_KWARGS

    @staticmethod
    def _interpret_columns_param(columns, none_error=False, none_columns=None):
        """Interprets the value provided to the columns parameter and returns
        a list version of it - if needed - and a string representation of it.
        """
        if columns is None:
            if none_error:
                raise ValueError((
                    'None is not a valid argument for the columns parameter of'
                    ' this pipeline stage.'))
            return ColumnsBasedPipelineStage._interpret_columns_param(
                columns=none_columns)
        if isinstance(columns, str):
            # always check str first, because it has __iter__
            return [columns], columns
        if callable(columns):
            # if isinstance(columns, ColumnQualifier):
            #     return columns, columns.__repr__() or ''
            return columns, columns.__doc__ or ''
        # if it was a single string it was already made a list, and it's not a
        # callable, so it's either an iterable of labels... or
        if hasattr(columns, '__iter__'):
            return columns, ', '.join(str(elem) for elem in columns)
        # a single non-string label.
        return [columns], str(columns)

    def __init__(
            self, columns, exclude_columns=None, desc_temp=None,
            none_columns='error', **kwargs):
        self._exclude_columns = exclude_columns
        if exclude_columns:
            self._exclude_columns, self._exc_col_str = \
                self._interpret_columns_param(exclude_columns)
        self._none_error = False
        self._none_cols = None
        # handle none_columns
        if isinstance(none_columns, str):
            if none_columns == 'error':
                self._none_error = True
            elif none_columns == 'all':
                self._none_cols = AllColumns()
            else:
                raise ValueError((
                    "'error' and 'all' are the only valid string arguments"
                    " to the none_columns constructor parameter!"))
        elif hasattr(none_columns, '__iter__'):
            self._none_cols = none_columns
        elif callable(none_columns):
            self._none_cols = none_columns
        else:
            raise ValueError((
                "Valid arguments to the none_columns constructor parameter"
                " are 'error', 'all', an iterable of labels or a callable!"
            ))
        # done handling none_columns
        self._col_arg, self._col_str = self._interpret_columns_param(
            columns, self._none_error, none_columns=self._none_cols)
        if exclude_columns:
            self._final_col_str = (
                f'{self._col_str} (except {self._exc_col_str})')
        else:
            self._final_col_str = f'{self._col_str}'
        if (kwargs.get('desc') is None) and desc_temp:
            kwargs['desc'] = desc_temp.format(self._final_col_str)
        if kwargs.get('exmsg') is None:
            kwargs['exmsg'] = (
                'Pipeline stage failed because not all columns {} '
                'were found in the input dataframe.'
            ).format(self._final_col_str)
        super().__init__(**kwargs)

    def _is_fittable(self):
        return is_fittable_column_qualifier(self._col_arg)

    @staticmethod
    def __get_cols_by_arg(col_arg, X, fit=False):
        try:
            if fit:
                # try to treat col_arg as a fittable column qualifier
                return col_arg.fit_transform(X)
            # else, no need to fit, so try to treat _col_arg as a callable
            return col_arg(X)
        except AttributeError:
            # got here cause col_arg has no fit_transform method...
            try:
                # so try and treat it as a callable again
                return col_arg(X)
            except TypeError:
                # calling col_arg 2 lines above failed; its a list of labels
                return col_arg
        except TypeError:
            # calling _col_arg 10 lines above failed; its a list of labels
            return col_arg

    def _get_columns(self, X, fit=False):
        cols = ColumnsBasedPipelineStage.__get_cols_by_arg(
            self._col_arg, X, fit=fit)
        if self._exclude_columns:
            exc_cols = ColumnsBasedPipelineStage.__get_cols_by_arg(
                self._exclude_columns, X, fit=fit)
            return [x for x in cols if x not in exc_cols]
        return cols

    def _prec(self, X, y=None):
        return set(self._get_columns(X)).issubset(X.columns)

    @abc.abstractmethod
    def _transformation(self, X, verbose, fit):
        raise NotImplementedError((
            "Classes extending ColumnsBasedPipelineStage must implement the "
            "_transformation method!"))

    def _fit_transform(self, X, verbose):
        self.is_fitted = True
        return self._transformation(X, verbose, fit=True)

    def _transform(self, X, verbose):
        return self._transformation(X, verbose, fit=False)


class AdHocStage(PdPipelineStage):
    """An ad-hoc stage of a pandas DataFrame-processing pipeline.

    The signature for both the `transform` and the optional `fit_transform`
    callables is adaptive: The first argument is used positionally (so no
    specific name is assumed or used) to supply the callable with the pandas
    DataFrame object to transform. The following additional keyword arguments
    are supplied if the are included in the callable's signature:
    `verbose` - Passed on from PdPipelineStage's `fit`, `fit_transform`
    and `apply` methods.

    `fit_context` and `application_context` - Provides fit-specific and
    application-specific contexts (see `PdpApplicationContext`) usually
    available to pipeline stages using `self.fit_context` and
    `self.application_context`.

    Parameters
    ----------
    transform : callable
        The transformation this stage applies to dataframes. If the
        fit_transform parameter is also populated than this transformation is
        only applied on calls to transform. See documentation for the exact
        signature.
    fit_transform : callable, optional
        The transformation this stage applies to dataframes, only on
        fit_transform. Optional. See documentation for the exact signature.
    prec : callable, default None
        A callable that returns a boolean value. Represent a a precondition
        used to determine whether this stage can be applied to a given
        dataframe. If None is given, set to a function always returning True.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[1, 'a'], [2, 'b']], [1, 2], ['num', 'char'])
    >>> drop_num = pdp.AdHocStage(
    ...   transform=lambda df: df.drop(['num'], axis=1),
    ...   prec=lambda df: 'num' in df.columns
    ... )
    >>> drop_num.apply(df)
      char
    1    a
    2    b
    """

    def __init__(self, transform, fit_transform=None, prec=None, **kwargs):
        if prec is None:
            prec = _always_true
        self._adhoc_transform = transform
        self._adhoc_fit_transform = fit_transform
        self._adhoc_prec = prec
        self._transform_kwargs = _get_args_list(self._adhoc_transform)
        try:
            self._fit_transform_kwargs = _get_args_list(
                self._adhoc_fit_transform)
        except TypeError:  # fit_transform is None
            self._fit_transform_kwargs = {}
        super().__init__(**kwargs)

    def _prec(self, X, y=None):
        try:
            return self._adhoc_prec(X, y)
        except TypeError as e:
            if len(PdPipelineStage._POS_ARG_MISMTCH_PAT.findall(str(e))) > 0:
                return self._adhoc_prec(X)
            raise e

    def _fit_transform(self, X, verbose):
        self.is_fitted = True
        if self._adhoc_fit_transform is None:
            self.is_fitted = True
            return self._transform(X, verbose)
        kwargs = {
            'verbose': verbose,
            'fit_context': self.fit_context,
            'application_context': self.application_context,
        }
        kwargs = {
            k: v for k, v in kwargs.items() if k in self._fit_transform_kwargs}
        return self._adhoc_fit_transform(X, **kwargs)

    def _transform(self, X, verbose):
        kwargs = {
            'verbose': verbose,
            'fit_context': self.fit_context,
            'application_context': self.application_context,
        }
        kwargs = {
            k: v for k, v in kwargs.items() if k in self._transform_kwargs}
        return self._adhoc_transform(X, **kwargs)


class PdPipeline(PdPipelineStage, collections.abc.Sequence):
    """A pipeline for processing pandas DataFrame objects.

    `transformer_getter` is useful to avoid applying pipeline stages that are
    aimed to filter out items in a big dataset to create a training set for a
    machine learning model, for example, but should not be applied on future
    individual items to be transformed by the fitted pipeline.

    Parameters
    ----------
    stages : list
        A list of PdPipelineStage objects making up this pipeline.
    transform_getter : callable, optional
        A callable that can be applied to the fitted pipeline to produce a
        sub-pipeline of it which should be used to transform dataframes after
        the pipeline has been fitted. If not given, the fitted pipeline is used
        entirely.
    """

    _DEF_EXC_MSG = 'Pipeline precondition failed!'

    def __init__(self, stages, transformer_getter=None, **kwargs):
        self._stages = stages
        self._trans_getter = transformer_getter
        self.is_fitted = False
        super_kwargs = {
            'exraise': False,
            'exmsg': PdPipeline._DEF_EXC_MSG,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    # implementing a collections.abc.Sequence abstract method
    def __getitem__(self, index):
        if isinstance(index, slice):
            return PdPipeline(self._stages[index])

        if isinstance(index, list) and all(isinstance(x, str) for x in index):
            stages = [stage for stage in self._stages if stage._name in index]
            return PdPipeline(stages)

        if isinstance(index, str):
            stages = [stage for stage in self._stages if stage._name == index]
            if len(stages) == 0:
                raise ValueError(f"'{index}' is not exist.")
            return stages[0]

        return self._stages[index]

    # implementing a collections.abc.Sequence abstract method
    def __len__(self):
        return len(self._stages)

    def _prec(self, X, y=None):
        # PdPipeline overrides apply in a way which makes this moot
        raise NotImplementedError

    def _post(self, X):
        # PdPipeline overrides apply in a way which makes this moot
        raise NotImplementedError

    def _transform(self, X, verbose):
        # PdPipeline overrides apply in a way which makes this moot
        raise NotImplementedError

    def _post_transform_lock(self):
        # Application context is discarded after pipeline application
        self.application_context = None
        self.fit_context.lock()

    def apply(
        self,
        X: pandas.DataFrame,
        y: Optional[pandas.Series] = None,
        exraise: Optional[bool] = None,
        verbose: Optional[bool] = False,
        time: Optional[bool] = False,
        fit_context: Optional[dict] = {},
        application_context: Optional[dict] = {},
    ):
        """Applies this pipeline stage to the given dataframe.

        If the stage is not fitted fit_transform is called. Otherwise,
        transform is called.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to which this pipeline stage will be applied.
        y : pandas.Series, optional
            A possible label column for processing of supervised learning
            datasets. Might also be transformed, if provided.
        exraise : bool, default None
            Determines behaviour if the precondition of composing stages is not
            fulfilled by the input dataframe: If True, a
            pdpipe.FailedPreconditionError is raised. If False, the stage is
            skipped. If not given, or set to None, the default behaviour of
            each stage is used, as determined by its 'exraise' constructor
            parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            is checked but before the application of the pipeline stage.
            Defaults to False.
        time : bool, default False
            If True, per-stage application time is measured and reported when
            pipeline application is done.
        fit_context : dict, option
            Context for the entire pipeline, is retained after the pipeline
            application is completed.
        application_context : dict, optional
            Context to add to the application context of this call. Can map
            str keys to arbitrary object values to be used by pipeline stages
            during this pipeline application. Discarded after pipeline
            application.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        if self.is_fitted:
            res = self.transform(
                X,
                y,
                exraise=exraise,
                verbose=verbose,
                time=time,
                application_context=application_context,
            )
            return res
        res = self.fit_transform(
            X,
            y,
            exraise=exraise,
            verbose=verbose,
            time=time,
            fit_context=fit_context,
            application_context=application_context,
        )
        return res

    def __timed_fit_transform(
        self,
        X: pandas.DataFrame,
        y: Optional[Iterable] = None,
        exraise: Optional[bool] = None,
        verbose: Optional[bool] = False,
        fit_context: Optional[dict] = {},
        application_context: Optional[dict] = {},
    ):
        self.fit_context = PdpApplicationContext()
        self.fit_context.update(fit_context)
        self.application_context = PdpApplicationContext()
        self.application_context.update(application_context)
        inter_X = X
        inter_y = y
        times = []
        prev = time.time()
        if y is None:
            for i, stage in enumerate(self._stages):
                try:
                    stage.fit_context = self.fit_context
                    stage.application_context = self.application_context
                    inter_X = stage.fit_transform(
                        X=inter_X,
                        y=None,
                        exraise=exraise,
                        verbose=verbose,
                    )
                    now = time.time()
                    times.append(now - prev)
                    prev = now
                except Exception as e:
                    raise PipelineApplicationError(
                        f"Exception raised in stage [ {i}] {stage}"
                    ) from e
        else:
            for i, stage in enumerate(self._stages):
                try:
                    stage.fit_context = self.fit_context
                    stage.application_context = self.application_context
                    inter_X, inter_y = stage.fit_transform(
                        X=inter_X,
                        y=inter_y,
                        exraise=exraise,
                        verbose=verbose,
                    )
                    now = time.time()
                    times.append(now - prev)
                    prev = now
                except Exception as e:
                    raise PipelineApplicationError(
                        f"Exception raised in stage [ {i}] {stage}"
                    ) from e
        self.is_fitted = True
        print("\nPipeline total application time: {:.3f}s.\n Details:".format(
            sum(times)))
        print(self.__times_str__(times))
        self._post_transform_lock()
        if y is None:
            return inter_X
        return inter_X, inter_y

    def fit_transform(
        self,
        X: pandas.DataFrame,
        y: Optional[Iterable] = None,
        exraise: Optional[bool] = None,
        verbose: Optional[bool] = False,
        time: Optional[bool] = False,
        fit_context: Optional[dict] = {},
        application_context: Optional[dict] = {},
    ):
        """Fits this pipeline and transforms the input dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to transform and fit this pipeline by.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Determines behaviour if the precondition of composing stages is not
            fulfilled by the input dataframe: If True, a
            pdpipe.FailedPreconditionError is raised. If False, the stage is
            skipped. If not given, or set to None, the default behaviour of
            each stage is used, as determined by its 'exraise' constructor
            parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            of each stage is checked but before its application. Otherwise, no
            messages are printed.
        time : bool, default False
            If True, per-stage application time is measured and reported when
            pipeline application is done.
        fit_context : dict, option
            Context for the entire pipeline, is retained after the pipeline
            application is completed.
        application_context : dict, optional
            Context to add to the application context of this call. Can map
            str keys to arbitrary object values to be used by pipeline stages
            during this pipeline application. Discarded after pipeline
            application.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        if time:
            return self.__timed_fit_transform(
                X, y, exraise=exraise,
                verbose=verbose,
                fit_context=fit_context,
                application_context=application_context)
        inter_X = X
        inter_y = y
        self.application_context = PdpApplicationContext()
        self.application_context.update(application_context)
        self.fit_context = PdpApplicationContext()
        self.fit_context.update(fit_context)
        if y is None:
            for i, stage in enumerate(self._stages):
                try:
                    stage.fit_context = self.fit_context
                    stage.application_context = self.application_context
                    inter_X = stage.fit_transform(
                        X=inter_X,
                        y=None,
                        exraise=exraise,
                        verbose=verbose,
                    )
                except Exception as e:
                    raise PipelineApplicationError(
                        f"Exception raised in stage [ {i}] {stage}"
                    ) from e
        else:
            for i, stage in enumerate(self._stages):
                try:
                    stage.fit_context = self.fit_context
                    stage.application_context = self.application_context
                    inter_X, inter_y = stage.fit_transform(
                        X=inter_X,
                        y=inter_y,
                        exraise=exraise,
                        verbose=verbose,
                    )
                except Exception as e:
                    raise PipelineApplicationError(
                        f"Exception raised in stage [ {i}] {stage}"
                    ) from e
        self._post_transform_lock()
        self.is_fitted = True
        if y is None:
            return inter_X
        return inter_X, inter_y

    def fit(
        self,
        X: pandas.DataFrame,
        y: Optional[Iterable] = None,
        exraise: Optional[bool] = None,
        verbose: Optional[bool] = False,
        time: Optional[bool] = False,
        fit_context: Optional[dict] = {},
        application_context: Optional[dict] = {},
    ):
        """Fits this pipeline without transforming the input dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to fit this pipeline by.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Determines behaviour if the precondition of composing stages is not
            fulfilled by the input dataframe: If True, a
            pdpipe.FailedPreconditionError is raised. If False, the stage is
            skipped. If not given, or set to None, the default behaviour of
            each stage is used, as determined by its 'exraise' constructor
            parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            of each stage is checked but before its application. Otherwise, no
            messages are printed.
        time : bool, default False
            If True, per-stage application time is measured and reported when
            pipeline application is done.
        fit_context : dict, option
            Context for the entire pipeline, is retained after the pipeline
            application is completed.
        application_context : dict, optional
            Context to add to the application context of this call. Can map
            str keys to arbitrary object values to be used by pipeline stages
            during this pipeline application.

        Returns
        -------
        pandas.DataFrame
            The input dataframe, unchanged.
        """
        self.fit_transform(
            X,
            y,
            exraise=exraise,
            verbose=verbose,
            time=time,
            fit_context=fit_context,
            application_context=application_context,
        )
        if y is None:
            return X
        return X, y

    def __timed_transform(
        self,
        X: pandas.DataFrame,
        y: Optional[Iterable[float]] = None,
        exraise: Optional[bool] = None,
        verbose: Optional[bool] = None,
        application_context: Optional[dict] = {},
    ) -> pandas.DataFrame:
        inter_X = X
        inter_y = y
        times = []
        prev = time.time()
        self.application_context = PdpApplicationContext()
        self.application_context.update(application_context)
        if y is None:
            for i, stage in enumerate(self._stages):
                try:
                    stage.fit_context = self.fit_context
                    stage.application_context = self.application_context
                    inter_X = stage.transform(
                        X=inter_X,
                        y=None,
                        exraise=exraise,
                        verbose=verbose,
                    )
                    now = time.time()
                    times.append(now - prev)
                    prev = now
                except Exception as e:
                    raise PipelineApplicationError(
                        f"Exception raised in stage [ {i}] {stage}"
                    ) from e
        else:
            for i, stage in enumerate(self._stages):
                try:
                    stage.fit_context = self.fit_context
                    stage.application_context = self.application_context
                    inter_X, inter_y = stage.transform(
                        X=inter_X,
                        y=inter_y,
                        exraise=exraise,
                        verbose=verbose,
                    )
                    now = time.time()
                    times.append(now - prev)
                    prev = now
                except Exception as e:
                    raise PipelineApplicationError(
                        f"Exception raised in stage [ {i}] {stage}"
                    ) from e
        self.is_fitted = True
        print("\nPipeline total application time: {:.3f}s.\n Details:".format(
            sum(times)))
        print(self.__times_str__(times))
        self._post_transform_lock()
        if y is None:
            return inter_X
        return inter_X, inter_y

    def transform(
        self,
        X: pandas.DataFrame,
        y: Optional[Iterable[float]] = None,
        exraise: Optional[bool] = None,
        verbose: Optional[bool] = None,
        time: Optional[bool] = False,
        application_context: Optional[dict] = {},
    ) -> pandas.DataFrame:
        """Transforms the given dataframe without fitting this pipeline.

        If any stage in this pipeline is fittable but is not fitted, an
        UnfittedPipelineStageError is raised before transformation starts.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to transform.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Determines behaviour if the precondition of composing stages is not
            fulfilled by the input dataframe: If True, a
            pdpipe.FailedPreconditionError is raised. If False, the stage is
            skipped. If not given, or set to None, the default behaviour of
            each stage is used, as determined by its 'exraise' constructor
            parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            of each stage is checked but before its application. Otherwise, no
            messages are printed.
        time : bool, default False
            If True, per-stage application time is measured and reported when
            pipeline application is done.
        application_context : dict, optional
            Context to add to the application context of this call. Can map
            str keys to arbitrary object values to be used by pipeline stages
            during this pipeline application.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        for stage in self._stages:
            if stage._is_fittable() and not stage.is_fitted:
                raise UnfittedPipelineStageError((
                    "PipelineStage {} in pipeline is fittable but"
                    " unfitted!").format(stage))
        if time:
            return self.__timed_transform(
                X, y, exraise=exraise, verbose=verbose,
                application_context=application_context)
        inter_X = X
        inter_y = y
        self.application_context = PdpApplicationContext()
        self.application_context.update(application_context)
        if y is None:
            for i, stage in enumerate(self._stages):
                try:
                    stage.fit_context = self.fit_context
                    stage.application_context = self.application_context
                    inter_X = stage.transform(
                        X=inter_X,
                        y=None,
                        exraise=exraise,
                        verbose=verbose,
                    )
                except Exception as e:
                    raise PipelineApplicationError(
                        f"Exception raised in stage [ {i}] {stage}"
                    ) from e
        else:
            for i, stage in enumerate(self._stages):
                try:
                    stage.fit_context = self.fit_context
                    stage.application_context = self.application_context
                    inter_X, inter_y = stage.transform(
                        X=inter_X,
                        y=inter_y,
                        exraise=exraise,
                        verbose=verbose,
                    )
                except Exception as e:
                    raise PipelineApplicationError(
                        f"Exception raised in stage [ {i}] {stage}"
                    ) from e
        self._post_transform_lock()
        if y is None:
            return inter_X
        return inter_X, inter_y

    __call__ = apply

    def __add__(self, other):
        if isinstance(other, PdPipeline):
            return PdPipeline([*self._stages, *other._stages])
        if isinstance(other, PdPipelineStage):
            return PdPipeline([*self._stages, other])
        return NotImplemented

    def __times_str__(self, times):
        res = "A pdpipe pipeline:\n"
        stime = sum(times)
        if stime > 0:  # pragma: no cover
            percentages = [100 * x / stime for x in times]
        else:  # pragma: no cover
            percentages = [0 for x in times]
        res += '[ 0] [{:0>5.2f}s ({:0>5.2f}%)]  '.format(
            times[0], percentages[0]
        ) + "\n      ".join(
            textwrap.wrap(self._stages[0].description())
        ) + '\n'
        for i, stage in enumerate(self._stages[1:]):
            res += '[{:>2}] [{:0>5.2f}s ({:0>5.2f}%)]  '.format(
                i + 1, times[i + 1], percentages[i + 1]
            ) + "\n      ".join(
                textwrap.wrap(stage.description())
            ) + '\n'
        return res

    def __str__(self):
        res = "A pdpipe pipeline:\n"
        res += '[ 0]  ' + "\n      ".join(
            textwrap.wrap(self._stages[0].description())) + '\n'
        for i, stage in enumerate(self._stages[1:]):
            res += '[{:>2}]  '.format(i + 1) + "\n      ".join(
                textwrap.wrap(stage.description())) + '\n'
        return res

    def _mem_str(self, total):
        total = asizeof(self)
        lines = []
        for i, stage in enumerate(self._stages):
            size = asizeof(stage)
            if size > 500000:  # pragma: no cover
                lines.append('[{:>2}] {:.2f}Mb ({:0>5.2f}%), {}\n'.format(
                    i, size / 1000000, 100 * size / total,
                    stage.description()))
            elif size > 1000:  # pragma: no cover
                lines.append('[{:>2}] {:.2f}Kb ({:0>5.2f}%), {}\n'.format(
                    i, size / 1000, 100 * size / total, stage.description()))
            else:
                lines.append('[{:>2}] {:}b ({:0>5.2f}%), {}\n'.format(
                    i, size, 100 * size / total, stage.description()))
            lines.append(stage._mem_str())
        return ''.join(lines)

    def memory_report(self):
        """Prints a detailed memory report of the pipeline object to screen.

        To get better memory estimates make sure the pympler Python package is
        installed. Without it, sys.getsizeof is used, which can be extremely
        underestimate memory size of Python objects.
        """
        print("=== Pipeline memory report ===")
        size = asizeof(self)
        if size > 500000:  # pragma: no cover
            print("Total pipeline size in memory: {:.2f}Mb".format(
                size / 1000000))
        elif size > 1000:  # pragma: no cover
            print("Total pipeline size in memory: {:.2f}Kb".format(
                size / 1000))
        else:
            print("Total pipeline size in memory: {:.2f}b".format(
                size))
        print("Per-stage memory structure:")
        print(self._mem_str(total=size))

    def get_transformer(self):
        """Return the transformer induced by this fitted pipeline.

           This transformer is a `pdpipe` pipeline that transforms input data
           in a way corresponding to this pipline after it has been fitted. By
           default this is the pipeline itself, but the `transform_getter`
           constructor parameter can be used to return a sub-pipeline of the
           fitted pipeline instead, for cases where some stages should only be
           applied when fitting this pipeline to data.

        Returns
        -------
        pdpipe.PdPipeline
            The corresponding transformer pipeline induced by this pipeline.
        """
        try:
            return self._trans_getter(self)
        except TypeError:  # pragma: no cover
            return self

    # def drop(self, index):
    #     """Returns this pipeline with the stage of the given index removed.
    #     Arguments
    #     ---------
    #     index


def make_pdpipeline(*stages: PdPipelineStage) -> PdPipeline:
    """Constructs a PdPipeline from the given pipeline stages.

    Parameters
    ----------
    *stages : pdpipe.PipelineStage objects
       PdPipeline stages given as positional arguments.

    Returns
    -------
    p : pdpipe.PdPipeline
        The resulting pipeline.

    Examples
    --------
    >>> import pdpipe as pdp
    >>> p = make_pdpipeline(pdp.ColDrop('count'), pdp.DropDuplicates())
    """
    return PdPipeline(stages=stages)



"""Handles for dynamic dataframe-method-wrapping pipeline stages."""

from pandas import DataFrame

from ..fly import (
    drop_rows_where,
    keep_rows_where,
)
from .df_transformer import (
    _DfMethodTransformerHandle,
    _is_dataframe_transform,
)
from .bound_column_potential import (
    get_bound_column_potential_by_label,
    _BoundColumnPotential,
)


class _DfHandle():

    def __init__(self) -> None:
        for attr_name in dir(DataFrame):
            attr = getattr(DataFrame, attr_name)
            if _is_dataframe_transform(attr_name, attr):
                # print(f"Adding {attr_name} of {attr}")
                handle = _DfMethodTransformerHandle(
                    method_name=attr_name,
                    doc=attr.__doc__,
                )
                setattr(self, attr_name, handle)
        self.drop_rows_where = drop_rows_where
        self.keep_rows_where = keep_rows_where

    def __getitem__(self, label: object) -> _BoundColumnPotential:
        return get_bound_column_potential_by_label(label)



"""Custom tests of pdpipe conditions meant to run manually."""

import pandas as pd
import pdpipe as pdp


if __name__ == "__main__":
    df = pd.DataFrame([[1, 4], [4, 5], [1, 11]], [1, 2, 3], ['a', 'b'])
    pline = pdp.PdPipeline([
        pdp.FreqDrop(2, 'a', prec=pdp.cond.HasAllColumns(['x']))
    ])
    pline.apply(df, verbose=True)



"""Can help expand series transforms whitelist and blacklist occasionally."""


if __name__ == '__main__':
    from pandas import Series
    from pdpipe.df.series_transformer import _has_series_transform_doc
    from rich.console import Console
    console = Console()
    print = console.print
    ser = Series([1, 2, 3])
    transforms = []
    not_transforms = []
    for attr_name in dir(Series):
        attr = getattr(Series, attr_name)
        current_has_series_transform_doc = False
        if _has_series_transform_doc(attr_name, attr):
            # print(f"Potential series transform: {attr_name} of {attr}")
            method = getattr(ser, attr_name)
            try:
                res = method()
                if isinstance(res, Series):
                    current_has_series_transform_doc = True
            except Exception:
                pass
            try:
                res = method(1)
                if isinstance(res, Series):
                    current_has_series_transform_doc = True
            except Exception:
                pass
            try:
                res = method(ser)
                if isinstance(res, Series):
                    current_has_series_transform_doc = True
            except Exception:
                pass
            if current_has_series_transform_doc:
                transforms.append(attr_name)
                print(
                    f"[blue bold]\t{attr_name}[/blue bold] [green]is a "
                    "series transform[/green]")
            else:
                not_transforms.append(attr_name)
                print(
                    f"[blue bold]\t{attr_name}[/blue bold] [red]is not a "
                    "series transform[/red]")
    print("\n[blue bold]Transforms[/blue bold]")
    print(f"{transforms}")
    print("\n[blue bold]Potential Non-Transforms[/blue bold]")
    print(f"{not_transforms}")



"""Testing the ApplicationContextEnricher stage."""

import pytest
import pandas as pd

from pdpipe import (
    PdPipelineStage,
    PdPipeline,
)
from pdpipe.basic_stages import ApplicationContextEnricher
from pdpipe.exceptions import PipelineApplicationError


DF1 = pd.DataFrame({'a': ['a', 'b', 'c', 'd'], 'b': [5, 6, 7, 1]})


class ApplicationContextAsserter(PdPipelineStage):

    def __init__(self, **kwargs):
        self._enrichments = kwargs
        super_kwargs = {
            'desc': "Assert application context enrichments"
        }
        super().__init__(**super_kwargs)

    def _prec(self, df):
        return True

    def _transform(self, df, verbose):
        for k, v in self._enrichments.items():
            assert self.application_context[k] == v


def test_application_context_basic():
    df1 = DF1
    pline = PdPipeline([
        ApplicationContextEnricher(
            bsum=lambda df: df['b'].sum(),
            bmean=lambda df: df['b'].mean(),
            bdiff=lambda df, application_context:
                application_context['bsum'] - application_context['bmean'],
            d=5,
        ),
        ApplicationContextAsserter(
            bsum=df1['b'].sum(),
            bmean=df1['b'].mean(),
            bdiff=df1['b'].sum() - df1['b'].mean(),
            d=5,
        ),
    ])
    pline(df1)


def test_application_context_error():
    df1 = DF1
    pline = PdPipeline([
        ApplicationContextEnricher(
            asum=lambda df: (2 + df['a']).sum(),
        ),
    ])
    with pytest.raises(PipelineApplicationError):
        pline(df1)



"""Classes for sklearn integration.

Despite similar names, there is a difference between pdpipe PdPipeline and
sklearn.pipeline.Pipeline. PdPipeline can only chain transformers while
scikit-learn Pipeline objects can further include the final estimator to
provide additional methods such as `predict` and `predict_proba`.

This means that by itself, pdpipe PdPipeline does not integrate well with some
of scikit-learn utility classes such as sklearn.model_selection.GridSearchCV
compared to sklearn.pipeline.Pipeline.

This module resolves such integration issues. Refer to the notebooks folder of
the pdpipe repository for complete examples.
"""

from typing import Callable
from functools import update_wrapper

import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_is_fitted

from .core import PdPipeline


def _estimator_has(attr):
    """Check if we can delegate a method to the underlying estimator.

    Calling a prediction method will only be available if `refit=True`. In
    such case, we check first the fitted best estimator. If it is not
    fitted, we check the unfitted estimator.

    Checking the unfitted estimator allows to use `hasattr` on the `SearchCV`
    instance even before calling `fit`.
    """

    def check(self):
        # raise an AttributeError if `attr` does not exist
        getattr(self.estimator, attr)
        return True

    return check


class _AvailableIfDescriptor:  # pragma: no cover
    """Implements a conditional property using the descriptor protocol.

    Using this class to create a decorator will raise an ``AttributeError``
    if check(self) returns a falsey value. Note that if check raises an error
    this will also result in hasattr returning false.

    See https://docs.python.org/3/howto/descriptor.html for an explanation of
    descriptors.
    """

    def __init__(self, fn, check, attribute_name):
        self.fn = fn
        self.check = check
        self.attribute_name = attribute_name

        # update the docstring of the descriptor
        update_wrapper(self, fn)

    def __get__(self, obj, owner=None):
        attr_err = AttributeError(
            f"This {repr(owner.__name__)} has no attribute "
            f"{repr(self.attribute_name)}"
        )
        if obj is not None:
            # delegate only on instances, not the classes.
            # this is to allow access to the docstrings.
            if not self.check(obj):
                raise attr_err

            # lambda, but not partial, allows help() to work with
            # update_wrapper
            out = lambda *args, **kwargs: self.fn(obj, *args, **kwargs)  # noqa
        else:

            def fn(*args, **kwargs):
                if not self.check(args[0]):
                    raise attr_err
                return self.fn(*args, **kwargs)

            # This makes it possible to use the decorated method as an
            # unbound method,
            # for instance when monkeypatching.
            out = lambda *args, **kwargs: fn(*args, **kwargs)  # noqa
        # update the docstring of the returned function
        update_wrapper(out, self.fn)
        return out


def available_if(check):
    """An attribute that is available only if check returns a truthy value.

    Parameters
    ----------
    check : callable
        When passed the object with the decorated method, this should return
        a truthy value if the attribute is available, and either return False
        or raise an AttributeError if not available.
    """
    return lambda fn: _AvailableIfDescriptor(
        fn, check, attribute_name=fn.__name__)


class PdPipelineAndSklearnEstimator(BaseEstimator):
    """A PdPipeline object chained before an sklearn estimator object.

    This kind of object can also be used with sklearn's GridSearchCV.

    See the pipeline_and_model.ipynb notebook in the notebooks folder of the
    pdpipe repository for a tutorial on how to use this class.

    Parameters
    ----------
    pipeline : PdPipeline
        The preprocssing pipeline to connect.
    model : sklearn.base.BaseEstimator
        The model to connect to the pipeline.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> from pdpipe.skintegrate import PdPipelineAndSklearnEstimator;
    >>> from sklearn.linear_model import LogisticRegression;
    >>> DF2 = pd.DataFrame(
    ...    data=[['-1',0], ['-1',0], ['1',1], ['1',1]],
    ...    index=[1, 2, 3, 4],
    ...    columns=['feature1', 'target']
    ... )
    >>> all_x = DF2[['feature1']]
    >>> all_y = DF2['target']
    >>> mp = PdPipelineAndSklearnEstimator(
    ...    pipeline=pdp.ColumnDtypeEnforcer({'feature1': int}),
    ...    estimator=LogisticRegression()
    ... )
    >>> mp.fit(all_x, all_y)
    <PdPipeline -> LogisticRegression>
    >>> res = mp.predict(all_x)
    """

    def __init__(
        self,
        pipeline: PdPipeline,
        estimator: BaseEstimator,
    ):
        self.pipeline = pipeline
        self.estimator = estimator
        # if hasattr(estimator, "score"):
        #     def _passthrough_scorer(estimator, *args, **kwargs):
        #         """Function that wraps estimator.score"""
        #         return estimator.score(*args, **kwargs)
        #     self.score = _passthrough_scorer

    def __str__(self):
        try:
            return f"<PdPipeline -> {self._est_cls_name}>"
        except AttributeError:
            self._est_cls_name = type(self.estimator).__name__
            return self.__str__()

    def __repr__(self):
        return self.__str__()

    def score(self, X, y=None):
        if y is None:
            post_X = self.pipeline.transform(X)
            return self.estimator.score(X)
        y = pd.Series(y)
        y.index = X.index
        post_X, post_y = self.pipeline.transform(X, y)
        assert len(post_X) == len(post_y)
        return self.estimator.score(post_X.values, post_y.values)

    @property
    def _estimator_type(self):
        return self.estimator._estimator_type

    @property
    def classes_(self):
        """Class labels.
        Only available when the estimator is a classifier.
        """
        _estimator_has("classes_")(self)
        return self.estimator.classes_

    def fit(self, X, y):
        """A reference implementation of a fitting function.

        Parameters
        ----------
        X : pandas.DataFrame, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in
            regression).

        Returns
        -------
        self : object
            Returns self.
        """
        # X, y = check_X_y(X, y, accept_sparse=True)
        y = pd.Series(y)
        assert len(X) == len(y)
        y.index = X.index
        post_X, post_y = self.pipeline.fit_transform(X=X, y=y)
        if post_y is None:
            self.estimator.fit(X=post_X.values, y=None)
        else:
            assert len(post_X) == len(post_y)
            self.estimator.fit(X=post_X.values, y=post_y.values)
        self.is_fitted_ = True
        return self

    @available_if(_estimator_has("predict"))
    def predict(self, X):
        """ A reference implementation of a predicting function.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            Returns an array of ones.
            The predicted labels or values for `X` based on the estimator with
            the best found parameters.
        """
        # X = check_array(X, accept_sparse=True)
        check_is_fitted(self, 'is_fitted_')
        post_X = self.pipeline.transform(X=X)
        y_pred = self.estimator.predict(X=post_X.values)
        return y_pred

    @available_if(_estimator_has("predict_proba"))
    def predict_proba(self, X):
        """Call predict_proba on the estimator with the best found parameters.
        Only available if the underlying estimator supports
        ``predict_proba``.

        Parameters
        ----------
        X : indexable, length n_samples
            Must fulfill the input assumptions of the
            underlying estimator.

        Returns
        -------
        y_pred : ndarray of shape (n_samples,) or (n_samples, n_classes)
            Predicted class probabilities for `X` based on the estimator with
            the best found parameters. The order of the classes corresponds
            to that in the fitted attribute :term:`classes_`.
        """
        check_is_fitted(self, 'is_fitted_')
        post_X = self.pipeline.transform(X=X)
        y_pred = self.estimator.predict_proba(X=post_X.values)
        return y_pred

    @available_if(_estimator_has("predict_log_proba"))
    def predict_log_proba(self, X):
        """Call predict_log_proba on the estimator with the best found parameters.
        Only available if the underlying estimator supports
        ``predict_log_proba``.

        Parameters
        ----------
        X : indexable, length n_samples
            Must fulfill the input assumptions of the
            underlying estimator.

        Returns
        -------
        y_pred : ndarray of shape (n_samples,) or (n_samples, n_classes)
            Predicted class log-probabilities for `X` based on the estimator
            with the best found parameters. The order of the classes
            corresponds to that in the fitted attribute :term:`classes_`.
        """
        check_is_fitted(self, 'is_fitted_')
        post_X = self.pipeline.transform(X=X)
        y_pred = self.estimator.predict_log_proba(X=post_X.values)
        return y_pred

    @available_if(_estimator_has("decision_function"))
    def decision_function(self, X):
        """Call decision_function on the estimator with the best found parameters.
        Only available if the underlying estimator supports
        ``decision_function``.

        Parameters
        ----------
        X : indexable, length n_samples
            Must fulfill the input assumptions of the
            underlying estimator.

        Returns
        -------
        y_score : ndarray of shape (n_samples,) or (n_samples, n_classes) \
                or (n_samples, n_classes * (n_classes-1) / 2)
            Result of the decision function for `X` based on the estimator with
            the best found parameters.
        """
        check_is_fitted(self, 'is_fitted_')
        post_X = self.pipeline.transform(X=X)
        y_score = self.estimator.decision_function(X=post_X.values)
        return y_score


class _PdPipeScorer:
    """A pdpipe scorer object wrapping a standard sklearn scorer.

    Parameters
    ----------
    scorer : Callable
        The wrapped sklearn scorer.
    """

    def __init__(self, scorer: Callable) -> None:
        self._scorer = scorer

    def __call__(
        self,
        estimator: PdPipelineAndSklearnEstimator,
        X: pd.DataFrame,
        y=None,
        **kwargs,
    ):
        post_X = estimator.pipeline.transform(X)
        return self._scorer(
            estimator.estimator,
            post_X,
            y,
            **kwargs,
        )

    def __repr__(self) -> str:
        rs = repr(self._scorer)
        return f'<PdPipeScorer: {rs}>'


def pdpipe_scorer_from_sklearn_scorer(scorer: Callable) -> Callable:
    """Converts an sklearn scorer to one that will work with pdpipe.

    The returned scorer function can then be used with sklearn's
    model-evaluation tools using cross-validation (such as
    model_selection.cross_val_score and model_selection.GridSearchCV), when
    searching over the hyperparameter space of a PdPipelineAndSklearnEstimator
    object.

    See the pipeline_and_model_with_test_test.ipynb notebook in the notebooks
    folder of the pdpipe repository for a complete example.

    Parameters
    ----------
    scorer : callable
        A function with the signature `scorer(estimator, X, y)`. To build one
        from an sklearn `score` function (with a signature of the form
        `score(y_true, y_pred, ...)`) use the `sklearn.metrics.make_scorer`
        function.

    Returns
    -------
    pdpipe_scorer : callable
        A scorer that is aware of the fact that PdPipelineAndSklearnEstimator
        has an inner pipeline object that should be used to transform input
        X (which is a dataframe when using pdpipe, and not a numpy.ndarray).
    """
    return _PdPipeScorer(scorer)



"""Testing pdpipe's df module."""

import pandas as pd

import pdpipe as pdp
from pdpipe.df.df_transformer import _DataFrameMethodTransformer
from pdpipe import df


def _test_df():
    return pd.DataFrame(
        data=[[1, 2, 'a'], [2, 4, 'b']],
        index=[1, 2],
        columns=['num1', 'num2', 'char']
    )


def test_df_set_axis():
    """Testing the ColDrop pipeline stage."""
    df = _test_df()  # noqa: F811
    assert 'num1' in df.columns
    stage = pdp.df.set_index(keys='num2')
    res_df = stage.apply(df)
    assert 'num1' in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns
    assert list(res_df.index) == [2, 4]

    pline = pdp.PdPipeline([
        pdp.df.set_index(keys='num2'),
        pdp.ColDrop('char'),
    ])
    stage = pline['set_index']
    assert isinstance(stage, _DataFrameMethodTransformer)
    res_df = pline(df)
    assert 'num1' in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' not in res_df.columns
    assert list(res_df.index) == [2, 4]

    # check inplace is ignored correctly
    df = _test_df()
    assert 'num1' in df.columns
    stage = pdp.df.set_index(keys='num2', drop=True, inplace=True)
    res_df = stage.apply(df)
    assert 'num1' in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns
    assert list(res_df.index) == [2, 4]
    assert 'num2' in df


def get_pipeline() -> pdp.PdPipeline:
    return pdp.PdPipeline([
        df.drop('char', axis=1),
        df.set_index(keys='num2'),
    ])


def test_more_df():
    df = _test_df()
    pline = get_pipeline()
    res_df = pline(df)
    assert 'num1' in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' not in res_df.columns
    assert list(res_df.index) == [2, 4]



"""Row qualifiers for pdpipe."""

from typing import List, Set, Union
from numbers import Number

import pandas


class RowQualifier(object):
    """An object that returns a boolean series from input dataframes.

    The boolean series will be of the length of the input dataframe, and so can
    be used as a boolean index to get a subset of the input dataframe's rows.

    Parameters
    ----------
    func : callable
        The function to apply to input dataframes.
    """

    def __init__(self, func: callable) -> None:
        self._rqfunc = func

    def __call__(self, X: pandas.DataFrame) -> pandas.Series:
        return self._rqfunc(X)

    def __repr__(self):
        if self._rqfunc.__doc__:  # pragma: no cover
            return f"<RowQualifier: Qualify rows with {self._rqfunc.__doc__}>"
        return "<RowQualifier: Qualify rows by function>"

    # --- overriding boolean operators ---

    class _AndQualifierFunc(object):
        """A pickle-able AND qualifier callable."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, X):
            return self.first(X) & self.second(X)

    def __and__(self, other):
        try:
            res_func = RowQualifier._AndQualifierFunc(
                first=self._rqfunc,
                second=other._rqfunc,
            )
            res_func.__doc__ = (
                f"{self._rqfunc.__doc__ or 'Anonymous qualifier 1'} AND "
                f"{other._rqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return RowQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _OrQualifierFunc(object):
        """A pickle-able OR qualifier callable."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, X):
            return self.first(X) | self.second(X)

    def __or__(self, other):
        try:
            res_func = RowQualifier._OrQualifierFunc(
                first=self._rqfunc,
                second=other._rqfunc,
            )
            res_func.__doc__ = (
                f"{self._rqfunc.__doc__ or 'Anonymous qualifier 1'} OR "
                f"{other._rqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return RowQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _XorQualifierFunc(object):
        """A pickle-able OR qualifier callable."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, X):
            return self.first(X) ^ self.second(X)

    def __xor__(self, other):
        try:
            res_func = RowQualifier._XorQualifierFunc(
                first=self._rqfunc,
                second=other._rqfunc,
            )
            res_func.__doc__ = (
                f"{self._rqfunc.__doc__ or 'Anonymous qualifier 1'} XOR "
                f"{other._rqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return RowQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _NotQualifierFunc(object):
        """A pickle-able NOT row qualifier callable."""

        def __init__(self, rq):
            self.rq = rq

        def __call__(self, X):
            return ~ self.rq(X)

    def __invert__(self):
        res_func = RowQualifier._NotQualifierFunc(rq=self._rqfunc)
        res_func.__doc__ = (
            f"NOT {self._rqfunc.__doc__ or 'Anonymous qualifier'}"
        )
        return RowQualifier(func=res_func)


class ColValGt(RowQualifier):
    """A row qualifier that qualifies rows with a value greater than a value.

    Parameters
    ----------
    label : object
        The label of the column the qualifier checks.
    value : Number
        The value to check against.
    """

    class _GtRowFunc(object):
        """A pickle-able gt callable class."""

        def __init__(self, label: object, value: Number) -> None:
            self.label = label
            self.value = value
            self.__doc__ = f"X[{label}] > {value}"

        def __call__(self, X: pandas.DataFrame) -> pandas.Series:
            return X[self.label].gt(self.value)

    def __init__(self, label: object, value: Number) -> None:
        super().__init__(
            func=ColValGt._GtRowFunc(label, value),
        )


class ColValGe(RowQualifier):
    """A row qualifier that qualifies rows with a value greater than or equal
    to a value.

    Parameters
    ----------
    label : object
        The label of the column the qualifier checks.
    value : Number
        The value to check against.
    """

    class _GeRowFunc(object):
        """A pickle-able ge callable class."""

        def __init__(self, label: object, value: Number) -> None:
            self.label = label
            self.value = value
            self.__doc__ = f"X[{label}] => {value}"

        def __call__(self, X: pandas.DataFrame) -> pandas.Series:
            return X[self.label].ge(self.value)

    def __init__(self, label: object, value: Number) -> None:
        super().__init__(
            func=ColValGe._GeRowFunc(label, value),
        )


class ColValLt(RowQualifier):
    """A row qualifier that qualifies rows with a value lesser than a value.

    Parameters
    ----------
    label : object
        The label of the column the qualifier checks.
    value : Number
        The value to check against.
    """

    class _LtRowFunc(object):
        """A pickle-able lt callable class."""

        def __init__(self, label: object, value: Number) -> None:
            self.label = label
            self.value = value
            self.__doc__ = f"X[{label}] < {value}"

        def __call__(self, X: pandas.DataFrame) -> pandas.Series:
            return X[self.label].lt(self.value)

    def __init__(self, label: object, value: Number) -> None:
        super().__init__(
            func=ColValLt._LtRowFunc(label, value),
        )


class ColValLe(RowQualifier):
    """A row qualifier that qualifies rows with a value lesser than or equal
    to a value.

    Parameters
    ----------
    label : object
        The label of the column the qualifier checks.
    value : Number
        The value to check against.
    """

    class _LeRowFunc(object):
        """A pickle-able le callable class."""

        def __init__(self, label: object, value: Number) -> None:
            self.label = label
            self.value = value
            self.__doc__ = f"X[{label}] <= {value}"

        def __call__(self, X: pandas.DataFrame) -> pandas.Series:
            return X[self.label].le(self.value)

    def __init__(self, label: object, value: Number) -> None:
        super().__init__(
            func=ColValLe._LeRowFunc(label, value),
        )


class ColValEq(RowQualifier):
    """A row qualifier that qualifies rows with a value equal to a value.

    Parameters
    ----------
    label : object
        The label of the column the qualifier checks.
    value : Number
        The value to check against.
    """

    class _EqRowFunc(object):
        """A pickle-able eq callable class."""

        def __init__(self, label: object, value: Number) -> None:
            self.label = label
            self.value = value
            self.__doc__ = f"X[{label}] == {value}"

        def __call__(self, X: pandas.DataFrame) -> pandas.Series:
            return X[self.label].eq(self.value)

    def __init__(self, label: object, value: Number) -> None:
        super().__init__(
            func=ColValEq._EqRowFunc(label, value),
        )


class ColValNe(RowQualifier):
    """A row qualifier that qualifies rows with a value not equal to a value.

    Parameters
    ----------
    label : object
        The label of the column the qualifier checks.
    value : Number
        The value to check against.
    """

    class _NeRowFunc(object):
        """A pickle-able ne callable class."""

        def __init__(self, label: object, value: Number) -> None:
            self.label = label
            self.value = value
            self.__doc__ = f"X[{label}] != {value}"

        def __call__(self, X: pandas.DataFrame) -> pandas.Series:
            return X[self.label].ne(self.value)

    def __init__(self, label: object, value: Number) -> None:
        super().__init__(
            func=ColValNe._NeRowFunc(label, value),
        )


class ColValIsIn(RowQualifier):
    """A row qualifier that qualifies rows with a value not equal to a value.

    Parameters
    ----------
    label : object
        The label of the column the qualifier checks.
    value_list : list of object
        The list of values to check against.
    """

    class _IsInRowFunc(object):
        """A pickle-able isin callable class."""

        def __init__(
            self,
            label: object,
            value_list: Union[List[object], Set[object]],
        ) -> None:
            self.label = label
            self.value_list = value_list
            self.__doc__ = f"X[{label}] is in {value_list}"

        def __call__(self, X: pandas.DataFrame) -> pandas.Series:
            return X[self.label].isin(self.value_list)

    def __init__(
        self,
        label: object,
        value_list: Union[List[object], Set[object]],
    ) -> None:
        super().__init__(
            func=ColValIsIn._IsInRowFunc(label, value_list),
        )


class ColValIsNa(RowQualifier):
    """A row qualifier that qualifies rows with a null value in a column.

    Parameters
    ----------
    label : object
        The label of the column the qualifier checks.
    """

    class _IsNaRowFunc(object):
        """A pickle-able isna callable class."""

        def __init__(
            self,
            label: object,
        ) -> None:
            self.label = label
            self.__doc__ = f"X[{label}] is NA"

        def __call__(self, X: pandas.DataFrame) -> pandas.Series:
            return X[self.label].isna()

    def __init__(
        self,
        label: object,
    ) -> None:
        super().__init__(func=ColValIsNa._IsNaRowFunc(label))


class ColValNotNa(RowQualifier):
    """A row qualifier that qualifies rows with a non-NA value in a column.

    Parameters
    ----------
    label : object
        The label of the column the qualifier checks.
    """

    class _NotNaRowFunc(object):
        """A pickle-able notna callable class."""

        def __init__(
            self,
            label: object,
        ) -> None:
            self.label = label
            self.__doc__ = f"X[{label}] is not NA"

        def __call__(self, X: pandas.DataFrame) -> pandas.Series:
            return X[self.label].notna()

    def __init__(
        self,
        label: object,
    ) -> None:
        super().__init__(func=ColValNotNa._NotNaRowFunc(label))


del pandas
del Number



"""Testing pdp.fly."""

import pytest
import pandas as pd
from pdpipe.fly import (
    drop_rows_where,
    keep_rows_where,
)


#     a   b
# 1   1   4
# 2   4   5
# 3  18  11
def _df1():
    return pd.DataFrame([[1, 4], [4, 5], [18, 11]], [1, 2, 3], ['a', 'b'])


#        a   b
# 1      1   4
# 2   Null   5
# 3     18  11
def _df2():
    return pd.DataFrame([[1, 4], [None, 5], [18, 11]], [1, 2, 3], ['a', 'b'])


def test_drop_where():
    df = _df1()
    res = (drop_rows_where['a'] > 4)(df)
    assert (res.index == [1, 2]).all()
    res = (drop_rows_where['a'] >= 4)(df)
    assert (res.index == [1]).all()
    res = (drop_rows_where['a'] < 4)(df)
    assert (res.index == [2, 3]).all()
    res = (drop_rows_where['a'] <= 4)(df)
    assert (res.index == [3]).all()
    res = (drop_rows_where['a'] == 4)(df)
    assert (res.index == [1, 3]).all()
    res = (drop_rows_where['a'] != 4)(df)
    assert (res.index == [2]).all()
    res = (drop_rows_where['a'].isin([1, 3, 5, 18]))(df)
    assert (res.index == [2]).all()
    df2 = _df2()
    res = (drop_rows_where['a'].isna())(df2)
    assert (res.index == [1, 3]).all()
    res = (drop_rows_where['a'].notna())(df2)
    assert (res.index == [2]).all()

    # unary operators
    res = (~(drop_rows_where['a'] > 4))(df)
    assert (res.index == [3]).all()

    # binary operators
    res = ((drop_rows_where['a'] <= 1) | (drop_rows_where['b'] >= 11))(df)
    assert (res.index == [2]).all()
    res = ((drop_rows_where['a'] <= 1) & (drop_rows_where['b'] >= 11))(df)
    assert (res.index == [1, 2, 3]).all()
    res = ((drop_rows_where['a'] <= 4) ^ (drop_rows_where['b'] >= 5))(df)
    assert (res.index == [2]).all()

    # errors
    dstage = drop_rows_where['a'] > 4
    with pytest.raises(TypeError):
        dstage & '4'
    with pytest.raises(TypeError):
        dstage | '4'
    with pytest.raises(TypeError):
        dstage ^ '4'
    with pytest.raises(ValueError):
        drop_rows_where['x':'y'] == 4

    # verbose
    stage = drop_rows_where['a'] > 4
    res = stage.apply(df, verbose=True)
    assert (res.index == [1, 2]).all()


def test_keep_where():
    df = _df1()
    res = (keep_rows_where['a'] > 4)(df)
    assert (res.index == [3]).all()
    res = (keep_rows_where['a'] >= 4)(df)
    assert (res.index == [2, 3]).all()
    res = (keep_rows_where['a'] < 4)(df)
    assert (res.index == [1]).all()
    res = (keep_rows_where['a'] <= 4)(df)
    assert (res.index == [1, 2]).all()
    res = (keep_rows_where['a'] == 4)(df)
    assert (res.index == [2]).all()
    res = (keep_rows_where['a'] != 4)(df)
    assert (res.index == [1, 3]).all()
    res = (keep_rows_where['a'].isin([1, 3, 5, 18]))(df)
    assert (res.index == [1, 3]).all()
    df2 = _df2()
    res = (keep_rows_where['a'].isna())(df2)
    assert (res.index == [2]).all()
    res = (keep_rows_where['a'].notna())(df2)
    assert (res.index == [1, 3]).all()

    # unary operators
    res = (~(keep_rows_where['a'] > 4))(df)
    assert (res.index == [1, 2]).all()

    # binary operators
    res = ((keep_rows_where['a'] <= 1) | (keep_rows_where['b'] >= 11))(df)
    assert (res.index == [1, 3]).all()
    res = ((keep_rows_where['a'] <= 1) & (keep_rows_where['b'] >= 11))(df)
    assert (res.index == []).all()
    res = ((keep_rows_where['a'] <= 4) ^ (keep_rows_where['b'] >= 5))(df)
    assert (res.index == [1, 3]).all()

    # errors
    dstage = keep_rows_where['a'] > 4
    with pytest.raises(TypeError):
        dstage & '4'
    with pytest.raises(TypeError):
        dstage | '4'
    with pytest.raises(TypeError):
        dstage ^ '4'
    with pytest.raises(ValueError):
        keep_rows_where['x':'y'] == 4

    # verbose
    stage = keep_rows_where['a'] > 4
    res = stage.apply(df, verbose=True)
    assert (res.index == [3]).all()



"""Shared testing functionalities for pdpipe."""

import os
import random


def random_pickle_path(pdpipe_tests_dpath: str) -> str:
    n = random.randint(1, 999999)
    return os.path.join(
        pdpipe_tests_dpath,
        f'pdpipe_pickled_obj_{n}',
    )
